"""A module for interfaces"""

import os

import numpy as np
from nipype.interfaces.base import File, SimpleInterface, TraitedSpec, traits
from nipype.utils.filemanip import fname_presuffix


class ClipInputSpec(TraitedSpec):
    in_file = File(exists=True, mandatory=True, desc='Input imaging file')
    out_file = File(desc='Output file name')
    minimum = traits.Float(
        -np.inf, usedefault=True, desc='Values under minimum are set to minimum'
    )
    maximum = traits.Float(np.inf, usedefault=True, desc='Values over maximum are set to maximum')


class ClipOutputSpec(TraitedSpec):
    out_file = File(desc='Output file name')


class Clip(SimpleInterface):
    """Simple clipping interface that clips values to specified minimum/maximum
    If no values are outside the bounds, nothing is done and the in_file is passed
    as the out_file without copying.
    """

    input_spec = ClipInputSpec
    output_spec = ClipOutputSpec

    def _run_interface(self, runtime):
        import nibabel as nb

        img = nb.load(self.inputs.in_file)
        data = img.get_fdata()

        out_file = self.inputs.out_file
        if out_file:
            out_file = os.path.join(runtime.cwd, out_file)

        if np.any((data < self.inputs.minimum) | (data > self.inputs.maximum)):
            if not out_file:
                out_file = fname_presuffix(
                    self.inputs.in_file, suffix='_clipped', newpath=runtime.cwd
                )
            np.clip(data, self.inputs.minimum, self.inputs.maximum, out=data)
            img.__class__(data, img.affine, img.header).to_filename(out_file)
        elif not out_file:
            out_file = self.inputs.in_file

        self._results['out_file'] = out_file
        return runtime


class Label2MaskInputSpec(TraitedSpec):
    in_file = File(exists=True, mandatory=True, desc='Input label file')
    label_val = traits.Int(mandatory=True, dec='Label value to create mask from')


class Label2MaskOutputSpec(TraitedSpec):
    out_file = File(desc='Output file name')


class Label2Mask(SimpleInterface):
    """Create mask file for a label from a multi-label segmentation"""

    input_spec = Label2MaskInputSpec
    output_spec = Label2MaskOutputSpec

    def _run_interface(self, runtime):
        import nibabel as nb

        img = nb.load(self.inputs.in_file)

        mask = np.uint16(img.dataobj) == self.inputs.label_val
        out_img = img.__class__(mask, img.affine, img.header)
        out_img.set_data_dtype(np.uint8)

        out_file = fname_presuffix(self.inputs.in_file, suffix='_mask', newpath=runtime.cwd)

        out_img.to_filename(out_file)

        self._results['out_file'] = out_file
        return runtime



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

import os
import sys
from datetime import datetime, timezone

from packaging.version import Version, parse
from sphinx import __version__ as sphinxversion

import nibabies

# -- Path setup --------------------------------------------------------------
here = os.path.dirname(__file__)
# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
sys.path.append(os.path.join(here, 'sphinxext'))
sys.path.insert(0, os.path.join(here, '..', 'wrapper'))

# this is only available after sphinxext to PATH
from github_link import make_linkcode_resolve  # noqa: E402

# -- General configuration ---------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
needs_sphinx = '1.5.3'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.intersphinx',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.linkcode',
    'sphinx.ext.napoleon',
    'sphinxcontrib.bibtex',
    'sphinxarg.ext',  # argparse extension
    'nipype.sphinxext.plot_workflow',
    'myst_parser',  # allow markdown
    # 'sphinx-togglebutton',  # collapse admonitions
]

bibtex_bibfiles = ['../nibabies/data/boilerplate.bib']

autodoc_mock_imports = ['numpy', 'nibabel', 'nilearn']
if parse(sphinxversion) >= parse('1.7.0'):
    autodoc_mock_imports = [
        'pandas',
        'nilearn',
        'seaborn',
    ]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


source_suffix = ['.rst', '.md']

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'shibuya'

# Options specific to theme
html_theme_options = {
    'color_mode': 'light',
    'dark_code': True,
    'github_url': 'https://github.com/nipreps/nibabies',
    'nav_links': [
        {
            'title': 'NiPreps Homepage',
            'url': 'https://nipreps.org',
            'external': True,
        },
        {
            'title': 'Docker Hub',
            'url': 'https://hub.docker.com/r/nipreps/nibabies',
            'external': True,
        },
    ],
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# -- Napoleon parameters -----------------------------------------------------
# Accept custom section names to be parsed for numpy-style docstrings
# of parameters.
# Requires pinning sphinxcontrib-napoleon to a specific commit while
# https://github.com/sphinx-contrib/napoleon/pull/10 is merged.
napoleon_use_param = False
napoleon_custom_sections = [
    ('Inputs', 'Parameters'),
    ('Outputs', 'Parameters'),
]

# -- MyST parameters ---------------------------------------------------------

myst_heading_anchors = 3
myst_enable_extensions = [
    'colon_fence',
    'substitution',
]

linkcode_resolve = make_linkcode_resolve(
    'nibabies',
    'https://github.com/nipreps/' 'nibabies/blob/{revision}/' '{package}/{path}#L{lineno}',
)

project = 'NiBabies'
author = 'The NiPreps developers'

copyright = f'2021-{datetime.now(tz=timezone.utc)}, {author}'

nibabies_ver = Version(nibabies.__version__)
release = 'version' if nibabies_ver.is_prerelease else nibabies_ver.public

# to avoid Python highlighting in literal text
highlight_language = 'none'



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Within-baby registration of a T1w into a T2w image."""

from __future__ import annotations

from collections import defaultdict

from nipype.interfaces import (
    utility as niu,
)
from nipype.interfaces.ants.base import Info as ANTsInfo
from nipype.pipeline import engine as pe
from niworkflows.engine import Workflow, tag
from niworkflows.interfaces.fixes import FixHeaderApplyTransforms as ApplyTransforms
from smriprep.workflows.fit.registration import (
    TemplateDesc,
    _fmt_cohort,
    get_metadata,
    tf_ver,
)


@tag('anat.coreg')
def init_coregistration_wf(
    *,
    bspline_fitting_distance: int = 200,
    mem_gb: float = 3.0,
    omp_nthreads: int | None = None,
    sloppy: bool = False,
    debug: bool = False,
    t1w_mask: bool = False,
    probmap: bool = True,
    name: str = 'coregistration_wf',
):
    """
    Set-up a T2w-to-T1w within-baby co-registration framework.

    See the ANTs' registration config file (under ``nibabies/data``) for further
    details.
    The main surprise in it is that, for some participants, accurate registration
    requires extra degrees of freedom (one affine level and one SyN level) to ensure
    that the T1w and T2w images align well.
    I attribute this requirement to the following potential reasons:

      * The T1w image and the T2w image were acquired in different sessions, apart in
        time enough for growth to happen.
        Although this is, in theory possible, it doesn't seem the images we have tested
        on are acquired on different sessions.
      * The skull is still so malleable that a change of position of the baby inside the
        coil made an actual change on the overall shape of their head.
      * Nonlinear distortions of the T1w and T2w images are, for some reason, more notorious
        for babies than they are for adults.
        We would need to look into each sequence's details to confirm this.

    Parameters
    ----------
    bspline_fitting_distance : :obj:`float`
        Distance in mm between B-Spline control points for N4 INU estimation.
    mem_gb : :obj:`float`
        Base memory fingerprint unit.
    name : :obj:`str`
        This particular workflow's unique name (Nipype requirement).
    omp_nthreads : :obj:`int`
        The number of threads for individual processes in this workflow.
    sloppy : :obj:`bool`
        Run in *sloppy* mode.
    debug : :obj:`bool`
        Produce intermediate registration files
    t1w_mask : :obj:`bool`
        A precomputed mask for the T1w is available. In this case, generate a
        quick mask to assist in coregistration, but use the precomputed mask
        as the final output.
    probmap: :obj:`bool`
        A probabilistic brainmask is present in T2w space.

    Inputs
    ------
    in_t1w : :obj:`str`
        The preprocessed input T1w image (Denoising/INU/Clipping)
    in_t2w : :obj:`str`
        The preprocessed input T2w image (Denoising/INU/Clipping)
    in_mask : :obj:`str`
        The brainmask.
        If `t1w_mask` is False, will be in T2w space.
        If `t1w_mask` is True, will be in T1w space.
    in_probmap : :obj:`str`
        The probabilistic brainmask, as obtained in T2w space.

    Outputs
    -------
    t1w_coreg : :obj:`str`
        The preprocessed T1w image (INU and clipping), in its native space.
    t2w_coreg : :obj:`str`
        The preprocessed T2w image (INU and clipping), aligned into the T1w's space.
    t1w_brain : :obj:`str`
        The preprocessed, brain-extracted T1w image.
    t1w_mask : :obj:`str`
        The binary brainmask in T1w space.
    t1w2t2w_xfm : :obj:`str`
        The T1w-to-T2w mapping.

    """
    from nipype.interfaces.ants import N4BiasFieldCorrection
    from niworkflows.interfaces.fixes import FixHeaderApplyTransforms as ApplyTransforms
    from niworkflows.interfaces.fixes import FixHeaderRegistration as Registration
    from niworkflows.interfaces.nibabel import ApplyMask, Binarize, BinaryDilation

    from nibabies.utils.misc import get_file

    workflow = pe.Workflow(name)

    inputnode = pe.Node(
        niu.IdentityInterface(fields=['in_t1w', 'in_t2w', 'in_mask', 'in_probmap']),
        name='inputnode',
    )
    outputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                't1w_preproc',
                't1w_brain',
                't1w_mask',
                't1w2t2w_xfm',
                't2w2t1w_xfm',
                't2w_preproc',
            ]
        ),
        name='outputnode',
    )

    # Dilate t2w mask for easier t1->t2 registration
    fixed_masks_arg = pe.Node(niu.Merge(3), name='fixed_masks_arg', run_without_submitting=True)
    reg_mask = pe.Node(BinaryDilation(radius=8, iterations=3), name='reg_mask')
    refine_mask = pe.Node(BinaryDilation(radius=8, iterations=1), name='refine_mask')

    # Set up T1w -> T2w within-subject registration
    coreg = pe.Node(
        Registration(from_file=get_file('nibabies', 'data/t1-t2-coreg.json')),
        name='coreg',
        n_procs=omp_nthreads,
        mem_gb=mem_gb,
    )
    coreg.inputs.float = sloppy
    if debug:
        coreg.inputs.args = '--write-interval-volumes 5'
        coreg.inputs.output_inverse_warped_image = sloppy
        coreg.inputs.output_warped_image = sloppy

    final_n4 = pe.Node(
        N4BiasFieldCorrection(
            dimension=3,
            bspline_fitting_distance=bspline_fitting_distance,
            save_bias=True,
            copy_header=True,
            n_iterations=[50] * 5,
            convergence_threshold=1e-7,
            rescale_intensities=True,
            shrink_factor=4,
        ),
        n_procs=omp_nthreads,
        name='final_n4',
    )
    # Move the T2w into T1w space, and apply the mask to the T1w
    map_t2w = pe.Node(ApplyTransforms(interpolation='BSpline'), name='map_t2w', mem_gb=1)
    apply_mask = pe.Node(ApplyMask(), name='apply_mask')

    # fmt: off
    workflow.connect([
        (inputnode, final_n4, [('in_t1w', 'input_image')]),
        (inputnode, coreg, [('in_t1w', 'moving_image'),
                            ('in_t2w', 'fixed_image')]),
        (reg_mask, fixed_masks_arg, [
            ('out_file', 'in1'),
            ('out_file', 'in2')]),
        (refine_mask, fixed_masks_arg, [('out_file', 'in3')]),
        (inputnode, map_t2w, [
            ('in_t1w', 'reference_image'),
            ('in_t2w', 'input_image')]),
        (fixed_masks_arg, coreg, [('out', 'fixed_image_masks')]),
        (coreg, map_t2w, [
            ('inverse_composite_transform', 'transforms'),
        ]),
        (final_n4, apply_mask, [('output_image', 'in_file')]),
        (final_n4, outputnode, [('output_image', 't1w_preproc')]),
        (map_t2w, outputnode, [('output_image', 't2w_preproc')]),
        (apply_mask, outputnode, [('out_file', 't1w_brain')]),
        (coreg, outputnode, [('composite_transform', 't1w2t2w_xfm')]),
        (coreg, outputnode, [('inverse_composite_transform', 't2w2t1w_xfm')]),
    ])
    # fmt: on

    if t1w_mask:
        # The input mask is already in T1w space.
        # Generate a quick, rough mask of the T2w to be used to facilitate co-registration.
        from sdcflows.interfaces.brainmask import BrainExtraction

        masker = pe.Node(BrainExtraction(), name='t2w_masker')
        # fmt:off
        workflow.connect([
            (inputnode, masker, [('in_t2w', 'in_file')]),
            (masker, reg_mask, [('out_mask', 'in_file')]),
            (masker, refine_mask, [('out_mask', 'in_file')]),
            (inputnode, apply_mask, [('in_mask', 'in_mask')]),
            (inputnode, outputnode, [('in_mask', 't1w_mask')]),
        ])
        # fmt:on
        return workflow

    if probmap:
        # The T2w mask from the brain extraction workflow will be mapped to T1w space
        map_mask = pe.Node(ApplyTransforms(interpolation='Gaussian'), name='map_mask', mem_gb=1)
        thr_mask = pe.Node(Binarize(thresh_low=0.80), name='thr_mask')
        # fmt:off
        workflow.connect([
            (inputnode, reg_mask, [('in_mask', 'in_file')]),
            (inputnode, refine_mask, [('in_mask', 'in_file')]),
            (inputnode, map_mask, [
                ('in_t1w', 'reference_image'),
                ('in_probmap', 'input_image')]),
            (coreg, map_mask, [
                ('inverse_composite_transform', 'transforms'),
            ]),
            (map_mask, thr_mask, [('output_image', 'in_file')]),
            (map_mask, final_n4, [('output_image', 'weight_image')]),
            (thr_mask, outputnode, [('out_mask', 't1w_mask')]),
            (thr_mask, apply_mask, [('out_mask', 'in_mask')]),
        ])
        # fmt:on
        return workflow

    # A precomputed T2w mask was provided
    map_precomp_mask = pe.Node(
        ApplyTransforms(interpolation='MultiLabel'), name='map_precomp_mask'
    )
    # fmt:off
    workflow.connect([
        (inputnode, reg_mask, [('in_mask', 'in_file')]),
        (inputnode, refine_mask, [('in_mask', 'in_file')]),
        (inputnode, map_precomp_mask, [
            ('in_t1w', 'reference_image'),
            ('in_mask', 'input_image')]),
        (coreg, map_precomp_mask, [
            ('inverse_composite_transform', 'transforms'),
        ]),
        (map_precomp_mask, final_n4, [('output_image', 'weight_image')]),
        (map_precomp_mask, outputnode, [('output_image', 't1w_mask')]),
        (map_precomp_mask, apply_mask, [('output_image', 'in_mask')]),
    ])
    # fmt:on
    return workflow


@tag('anat.coreg-derivatives')
def init_coregister_derivatives_wf(
    *, t1w_mask: bool, t1w_aseg: bool, t2w_aseg: bool, name: str = 'coregister_derivatives_wf'
):
    """Move derivatives from T1w / T2w space."""
    workflow = pe.Workflow(name=name)
    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=['t1w_ref', 't2w_ref', 't1w2t2w_xfm', 't1w_mask', 't1w_aseg', 't2w_aseg']
        ),
        name='inputnode',
    )
    outputnode = pe.Node(
        niu.IdentityInterface(fields=['t2w_mask', 't1w_aseg', 't2w_aseg']), name='outputnode'
    )

    if t1w_mask:
        t1wmask2t2w = pe.Node(ApplyTransforms(interpolation='MultiLabel'), name='t1wmask2t2w')
        # fmt:off
        workflow.connect([
            (inputnode, t1wmask2t2w, [
                ('t1w_mask', 'input_image'),
                ('t1w2t2w_xfm', 'transforms'),
                ('t2w_ref', 'reference_image')]),
            (t1wmask2t2w, outputnode, [('output_image', 't2w_mask')])
        ])
        # fmt:on
    if t1w_aseg:
        # fmt:off
        t1waseg2t2w = pe.Node(ApplyTransforms(interpolation='MultiLabel'), name='t1waseg2t2w')
        workflow.connect([
            (inputnode, t1waseg2t2w, [
                ('t1w_aseg', 'input_image'),
                ('t1w2t2w_xfm', 'transforms'),
                ('t2w_ref', 'reference_image')]),
            (t1waseg2t2w, outputnode, [('output_image', 't2w_aseg')])
        ])
        # fmt:on
    if t2w_aseg:
        # fmt:off
        t2waseg2t1w = pe.Node(ApplyTransforms(interpolation='MultiLabel'), name='t2waseg2t1w')
        t2waseg2t1w.inputs.invert_transform_flags = [True, False]
        workflow.connect([
            (inputnode, t2waseg2t1w, [
                ('t2w_aseg', 'input_image'),
                ('t1w2t2w_xfm', 'transforms'),
                ('t1w_ref', 'reference_image')]),
            (t2waseg2t1w, outputnode, [('output_image', 't1w_aseg')])
        ])
        # fmt:on
    return workflow


@tag('anat.concat-reg')
def init_concat_registrations_wf(
    *,
    templates,
    name='concat_registrations_wf',
):
    """
    Concatenate two transforms to produce a single composite transform from native to template.

    Parameters
    ----------
    templates : :obj:`list` of :obj:`str`
        List of standard space fullnames (e.g., ``MNI152NLin6Asym``
        or ``MNIPediatricAsym:cohort-4``) which are targets for spatial
        normalization.

    Inputs
    ------
    anat_preproc
        The anatomical reference, to be used as a reference image for std2anat_xfm
    intermediate
        Standard space fullname (usually ``MNIInfant:cohort-X``) which serves as
        the intermediate space between native and *template*
    anat2std_xfm
        The incoming anat2std transform (from native to MNIInfant)
    std2anat_xfm
        The incoming std2anat transform (from MNIInfant to native)

    Outputs
    -------
    anat2std_xfm
        Anatomical -> MNIInfant -> template transform.
    std2anat_xfm
        The template -> MNIInfant -> anatomical transform.
    template
        Template name extracted from the input parameter ``template``, for further
        use in downstream nodes.
    template_spec
        Template specifications extracted from the input parameter ``template``, for
        further use in downstream nodes.

    """
    from nibabies.interfaces.download import RetrievePoochFiles
    from nibabies.interfaces.patches import CompositeTransformUtil

    ntpls = len(templates)
    workflow = Workflow(name=name)

    if templates:
        workflow.__desc__ = """\
Volume-based spatial normalization to {targets} ({targets_id}) was performed by
concatenating two registrations with `antsRegistration` (ANTs {ants_ver}). First, the
anatomical reference was registered to the Infant MNI templates (@mniinfant). Separately,
the Infant MNI templates were registered to {targets}, with the saved transform to template
stored for reuse and accessed with *TemplateFlow* [{tf_ver}, @templateflow]:
""".format(
            ants_ver=ANTsInfo.version() or '(version unknown)',
            targets='{} standard space{}'.format(
                defaultdict('several'.format, {1: 'one', 2: 'two', 3: 'three', 4: 'four'})[ntpls],
                's' * (ntpls != 1),
            ),
            targets_id=', '.join(templates),
            tf_ver=tf_ver,
        )

        # Append template citations to description
        for template in templates:
            template_meta = get_metadata(template.split(':')[0])
            template_refs = ['@{}'.format(template.split(':')[0].lower())]

            if template_meta.get('RRID', None):
                template_refs += [f'RRID:{template_meta["RRID"]}']

            workflow.__desc__ += """\
*{template_name}* [{template_refs}; TemplateFlow ID: {template}]""".format(
                template=template,
                template_name=template_meta['Name'],
                template_refs=', '.join(template_refs),
            )
            workflow.__desc__ += '.\n' if template == templates[-1] else ', '

    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'template',  # template identifier (name[+cohort])
                'intermediate',  # intermediate space (name[+cohort])
                'anat2int_xfm',  # anatomical -> intermediate
                'int2anat_xfm',  # intermediate -> anatomical
            ]
        ),
        name='inputnode',
    )
    inputnode.inputs.template = templates

    out_fields = [
        'anat2std_xfm',
        'std2anat_xfm',
        'template',
        'template_spec',
    ]
    outputnode = pe.Node(niu.IdentityInterface(fields=out_fields), name='outputnode')

    intermed_xfms = pe.MapNode(
        RetrievePoochFiles(),
        name='retrieve_xfms',
        iterfield=['target'],
        run_without_submitting=True,
    )

    split_desc = pe.MapNode(
        TemplateDesc(), run_without_submitting=True, iterfield='template', name='split_desc'
    )

    fmt_cohort = pe.MapNode(
        niu.Function(function=_fmt_cohort, output_names=['template', 'spec']),
        name='fmt_cohort',
        run_without_submitting=True,
        iterfield=['template', 'spec'],
    )

    # Disassemble each composite transform individually for readability
    dis_anat2int = pe.Node(
        CompositeTransformUtil(process='disassemble', output_prefix='anat2int'),
        name='dis_anat2int',
    )

    dis_int2std = pe.Node(
        CompositeTransformUtil(process='disassemble', output_prefix='int2std'),
        name='dis_int2std',
    )

    dis_std2int = pe.Node(
        CompositeTransformUtil(process='disassemble', output_prefix='std2int'),
        name='dis_std2int',
    )

    dis_int2anat = pe.Node(
        CompositeTransformUtil(process='disassemble', output_prefix='int2anat'),
        name='dis_int2anat',
    )

    order_anat2std = pe.Node(niu.Merge(4), name='order_anat2std', run_without_submitting=True)
    order_std2anat = pe.Node(niu.Merge(4), name='order_std2anat', run_without_submitting=True)

    assemble_anat2std = pe.Node(
        CompositeTransformUtil(process='assemble', out_file='anat2std.h5'),
        name='assemble_anat2std',
    )
    assemble_std2anat = pe.Node(
        CompositeTransformUtil(process='assemble', out_file='std2anat.h5'),
        name='assemble_std2anat',
    )

    workflow.connect([
        # Transform concatenation
        (inputnode, dis_anat2int, [('anat2int_xfm', 'in_file')]),
        (inputnode, dis_int2anat, [('int2anat_xfm', 'in_file')]),
        (inputnode, intermed_xfms, [('intermediate', 'intermediate'),
                                    ('template', 'target')]),
        (intermed_xfms, dis_int2std, [('int2tgt_xfm', 'in_file')]),
        (intermed_xfms, dis_std2int, [('tgt2int_xfm', 'in_file')]),
        (dis_anat2int, order_anat2std, [
            ('affine_transform', 'in1'),
            ('displacement_field', 'in2'),
        ]),
        (dis_int2std, order_anat2std, [
            ('affine_transform', 'in3'),
            ('displacement_field', 'in4'),
        ]),
        # Because std2anat are inverse transforms, warp is first
        (dis_std2int, order_std2anat, [
            ('affine_transform', 'in2'),
            ('displacement_field', 'in1'),
        ]),
        (dis_int2anat, order_std2anat, [
            ('affine_transform', 'in4'),
            ('displacement_field', 'in3'),
        ]),
        (order_anat2std, assemble_anat2std, [('out', 'in_file')]),
        (order_std2anat, assemble_std2anat, [('out', 'in_file')]),
        (assemble_anat2std, outputnode, [('out_file', 'anat2std_xfm')]),
        (assemble_std2anat, outputnode, [('out_file', 'std2anat_xfm')]),

        # Template name wrangling
        (inputnode, split_desc, [('template', 'template')]),
        (split_desc, fmt_cohort, [
            ('name', 'template'),
            ('spec', 'spec'),
        ]),
        (fmt_cohort, outputnode, [
            ('template', 'template'),
            ('spec', 'template_spec'),
        ]),
    ])  # fmt:skip

    return workflow



#!/usr/bin/env python
"""NiBabies runner."""

from .. import config


def main():
    """Entry point."""
    import gc
    import os
    import sys
    from pathlib import Path

    from ..utils.bids import write_bidsignore, write_derivative_description
    from .parser import parse_args
    from .workflow import build_boilerplate, build_workflow

    _cwd = os.getcwd()

    parse_args()

    if 'pdb' in config.execution.debug:
        from nibabies.utils.debug import setup_exceptionhook

        setup_exceptionhook()
        config.nipype.plugin = 'Linear'

    # collect and submit telemetry information
    # if `--notrack` is specified, nothing is done.
    if not config.execution.notrack and not config.execution.debug:
        from nibabies.utils.telemetry import setup_migas

        setup_migas()

    if 'participant' in config.workflow.analysis_level:
        _pool = None
        if config.nipype.plugin == 'MultiProc':
            import multiprocessing as mp
            from concurrent.futures import ProcessPoolExecutor
            from contextlib import suppress

            # should drastically reduce VMS
            # see https://github.com/nipreps/mriqc/pull/984 for more details
            os.environ['OMP_NUM_THREADS'] = '1'

            with suppress(RuntimeError):
                mp.set_start_method('fork')
            gc.collect()

            _pool = ProcessPoolExecutor(
                max_workers=config.nipype.nprocs,
                initializer=config._process_initializer,
                initargs=(_cwd, config.nipype.omp_nthreads),
            )

        config_file = config.execution.work_dir / config.execution.run_uuid / 'config.toml'
        config_file.parent.mkdir(exist_ok=True, parents=True)
        config.to_filename(config_file)

        # build the workflow within the same process
        # it still needs to be saved / loaded to be properly initialized
        retval = build_workflow(config_file)
        exitcode = retval['return_code']
        nibabies_wf = retval['workflow']

        # exit conditions:
        # - no workflow (--reports-only)
        # - retcode is not 0
        # - boilerplate only

        if nibabies_wf is None and not config.execution.reports_only:
            sys.exit(exitcode)

        if config.execution.write_graph:
            nibabies_wf.write_graph(graph2use='colored', format='svg', simple_form=True)

        if exitcode != 0:
            sys.exit(exitcode)

        # generate boilerplate
        build_boilerplate(nibabies_wf)
        if config.execution.boilerplate_only:
            sys.exit(exitcode)

        gc.collect()

        config.loggers.workflow.log(
            15,
            '\n'.join(['nibabies config:'] + [f'\t\t{s}' for s in config.dumps().splitlines()]),
        )
        config.loggers.workflow.log(25, 'nibabies started!')

        # Hack MultiProc's pool to reduce VMS
        _plugin = config.nipype.get_plugin()
        if _pool:
            from nipype.pipeline.plugins.multiproc import MultiProcPlugin

            multiproc = MultiProcPlugin(plugin_args=config.nipype.plugin_args)
            multiproc.pool = _pool
            _plugin = {'plugin': multiproc}

        gc.collect()
        try:
            nibabies_wf.run(**_plugin)
        except Exception as e:
            config.loggers.workflow.critical('nibabies failed: %s', e)
            raise
        else:
            config.loggers.workflow.log(25, 'nibabies finished successfully!')
            # Bother users with the boilerplate only iff the workflow went okay.
            boiler_file = config.execution.nibabies_dir / 'logs' / 'CITATION.md'
            if boiler_file.exists():
                if config.environment.exec_env in (
                    'singularity',
                    'docker',
                    'nibabies-docker',
                ):
                    boiler_file = Path('<OUTPUT_PATH>') / boiler_file.relative_to(
                        config.execution.output_dir
                    )
                config.loggers.workflow.log(
                    25,
                    'Works derived from this nibabies execution should include the '
                    f'boilerplate text found in {boiler_file}.',
                )

            if config.workflow.run_reconall:
                from niworkflows.utils.misc import _copy_any
                from templateflow import api

                dseg_tsv = str(
                    api.get(
                        'fsaverage',
                        hemi=None,
                        atlas=None,
                        segmentation='aparc',
                        suffix='dseg',
                        extension=['.tsv'],
                    )
                )
                _copy_any(dseg_tsv, str(config.execution.nibabies_dir / 'desc-aseg_dseg.tsv'))
                _copy_any(dseg_tsv, str(config.execution.nibabies_dir / 'desc-aparcaseg_dseg.tsv'))
        # errno = 0
        finally:
            from ..reports.core import generate_reports

            # Generate reports phase
            generate_reports(
                config.execution.unique_labels,
                config.execution.nibabies_dir,
                config.execution.run_uuid,
            )
            write_derivative_description(
                config.execution.bids_dir,
                config.execution.nibabies_dir,
                config.execution.dataset_links,
            )
            write_bidsignore(config.execution.nibabies_dir)


if __name__ == '__main__':
    raise RuntimeError(
        'Please `pip install` this and run via the commandline interfaces, `nibabies <command>`'
    )



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
"""Parser."""

from __future__ import annotations

import sys
import typing as ty

from .. import config

if ty.TYPE_CHECKING:
    from bids.layout import BIDSLayout


def _build_parser():
    """Build parser object."""
    from argparse import (
        Action,
        ArgumentDefaultsHelpFormatter,
        ArgumentParser,
        BooleanOptionalAction,
    )
    from functools import partial
    from pathlib import Path

    from niworkflows.utils.spaces import OutputReferencesAction, Reference
    from packaging.version import Version

    from .version import check_latest, is_flagged

    deprecations = {
        # parser attribute name: (replacement flag, version slated to be removed in)
        'clean_workdir',
        (None, '26.0.0'),
    }

    class DeprecatedAction(Action):
        def __call__(self, parser, namespace, values, option_string=None):
            new_opt, rem_vers = deprecations.get(self.dest, (None, None))
            msg = (
                f'{self.option_strings} has been deprecated and will be removed in '
                f'{rem_vers or "a later version"}.'
            )
            if new_opt:
                msg += f' Please use `{new_opt}` instead.'
            print(msg, file=sys.stderr)
            delattr(namespace, self.dest)

    class DerivToDict(Action):
        def __call__(self, parser, namespace, values, option_string=None):
            d = {}
            for spec in values:
                try:
                    name, loc = spec.split('=')
                    loc = Path(loc)
                except ValueError:
                    loc = Path(spec)
                    name = loc.name

                if name in d:
                    raise ValueError(f'Received duplicate derivative name: {name}')

                d[name] = loc
            setattr(namespace, self.dest, d)

    def _path_exists(path, parser):
        """Ensure a given path exists."""
        if path is None:
            raise parser.error('No value provided!')
        path = Path(path).absolute()
        if not path.exists():
            raise parser.error(f'Path does not exist: <{path}>.')
        return path

    def _dir_not_empty(path, parser):
        path = _path_exists(path, parser)
        if not path.is_dir():
            raise parser.error(f'Path is not a directory <{path}>.')
        for _ in path.iterdir():
            return path
        raise parser.error(f'Directory found with no contents <{path}>.')

    def _is_file(path, parser):
        """Ensure a given path exists and it is a file."""
        path = _path_exists(path, parser)
        if not path.is_file():
            raise parser.error(f'Path should point to a file (or symlink of file): <{path}>.')
        return path

    def _min_one(value, parser):
        """Ensure an argument is not lower than 1."""
        value = int(value)
        if value < 1:
            raise parser.error("Argument can't be less than one.")
        return value

    def _to_gb(value):
        scale = {'G': 1, 'T': 10**3, 'M': 1e-3, 'K': 1e-6, 'B': 1e-9}
        digits = ''.join([c for c in value if c.isdigit()])
        units = value[len(digits) :] or 'M'
        return int(digits) * scale[units[0]]

    def _drop_sub(value):
        return value[4:] if value.startswith('sub-') else value

    def _drop_ses(value):
        return value[4:] if value.startswith('ses-') else value

    def _process_value(value):
        import bids

        if value is None:
            return bids.layout.Query.NONE
        elif value == '*':
            return bids.layout.Query.ANY
        else:
            return value

    def _filter_pybids_none_any(dct):
        d = {}
        for k, v in dct.items():
            if isinstance(v, list):
                d[k] = [_process_value(val) for val in v]
            else:
                d[k] = _process_value(v)
        return d

    def _bids_filter(value, parser):
        from json import JSONDecodeError, loads

        if value:
            if Path(value).exists():
                try:
                    return loads(Path(value).read_text(), object_hook=_filter_pybids_none_any)
                except JSONDecodeError as e:
                    raise parser.error(f'JSON syntax error in: <{value}>.') from e
            else:
                raise parser.error(f'Path does not exist: <{value}>.')

    def _slice_time_ref(value, parser):
        if value == 'start':
            value = 0
        elif value == 'middle':
            value = 0.5
        try:
            value = float(value)
        except ValueError as e:
            raise parser.error(
                f"Slice time reference must be number, 'start', or 'middle'. Received {value}."
            ) from e
        if not 0 <= value <= 1:
            raise parser.error(f'Slice time reference must be in range 0-1. Received {value}.')
        return value

    def _str_none(val):
        if not isinstance(val, str):
            return val
        return None if val.lower() == 'none' else val

    verstr = f'NiBabies v{config.environment.version}'
    currentv = Version(config.environment.version)

    parser = ArgumentParser(
        description='NiBabies: Preprocessing workflows for infants v{config.environment.version}',
        formatter_class=ArgumentDefaultsHelpFormatter,
    )
    PathExists = partial(_path_exists, parser=parser)
    DirNotEmpty = partial(_dir_not_empty, parser=parser)
    IsFile = partial(_is_file, parser=parser)
    PositiveInt = partial(_min_one, parser=parser)
    BIDSFilter = partial(_bids_filter, parser=parser)
    SliceTimeRef = partial(_slice_time_ref, parser=parser)

    parser.description = f"""
NiBabies: Preprocessing workflows for infants v{config.environment.version}"""

    # Arguments as specified by BIDS-Apps
    # required, positional arguments
    # IMPORTANT: they must go directly with the parser object
    parser.add_argument(
        'bids_dir',
        action='store',
        type=PathExists,
        help='the root folder of a BIDS valid dataset (sub-XXXXX folders should '
        'be found at the top level in this folder).',
    )
    parser.add_argument(
        'output_dir',
        action='store',
        type=Path,
        help='the output path for the outcomes of preprocessing and visual reports',
    )
    parser.add_argument(
        'analysis_level',
        choices=['participant'],
        help='processing stage to be run, only "participant" in the case of '
        'NiBabies (see BIDS-Apps specification).',
    )

    # optional arguments
    parser.add_argument('--version', action='version', version=verstr)

    g_bids = parser.add_argument_group('Options for filtering BIDS queries')
    g_bids.add_argument(
        '--skip_bids_validation',
        '--skip-bids-validation',
        action='store_true',
        default=False,
        help='assume the input dataset is BIDS compliant and skip the validation',
    )
    g_bids.add_argument(
        '--participant-label',
        '--participant_label',
        action='store',
        nargs='+',
        type=_drop_sub,
        help='a space delimited list of participant identifiers or a single '
        'identifier (the sub- prefix can be removed)',
    )
    g_bids.add_argument(
        '-s',
        '--session-id',
        action='store',
        nargs='+',
        type=_drop_ses,
        help='a space delimited list of session identifiers or a single identifier',
    )
    # Re-enable when option is actually implemented
    # g_bids.add_argument('-r', '--run-id', action='store', default='single_run',
    #                     help='select a specific run to be processed')
    g_bids.add_argument(
        '-t', '--task-id', action='store', help='select a specific task to be processed'
    )
    g_bids.add_argument(
        '--echo-idx',
        action='store',
        type=int,
        help='select a specific echo to be processed in a multiecho series',
    )
    g_bids.add_argument(
        '--bids-filter-file',
        dest='bids_filters',
        action='store',
        type=BIDSFilter,
        metavar='FILE',
        help='a JSON file describing custom BIDS input filters using PyBIDS. '
        'For further details, please check out '
        'https://fmriprep.readthedocs.io/en/latest/faq.html#'
        'how-do-I-select-only-certain-files-to-be-input-to-fMRIPrep',
    )
    g_bids.add_argument(
        '-d',
        '--derivatives',
        action=DerivToDict,
        metavar='PACKAGE=PATH',
        type=str,
        nargs='+',
        help=(
            'Search PATH(s) for pre-computed derivatives. '
            'These may be provided as named folders '
            '(e.g., `--derivatives smriprep=/path/to/smriprep`).'
        ),
    )
    g_bids.add_argument(
        '--bids-database-dir',
        metavar='PATH',
        type=PathExists,
        help='Path to an existing PyBIDS database folder, for faster indexing '
        '(especially useful for large datasets).',
    )
    g_bids.add_argument(
        '--copy-derivatives',
        action='store_true',
        help='Copy any found derivatives into output directory',
    )

    g_perfm = parser.add_argument_group('Options to handle performance')
    g_perfm.add_argument(
        '--nprocs',
        '--nthreads',
        '--n_cpus',
        '--n-cpus',
        dest='nprocs',
        action='store',
        type=PositiveInt,
        help='maximum number of threads across all processes',
    )
    g_perfm.add_argument(
        '--omp-nthreads',
        action='store',
        type=PositiveInt,
        help='maximum number of threads per-process',
    )
    g_perfm.add_argument(
        '--mem',
        '--mem_mb',
        '--mem-mb',
        dest='memory_gb',
        action='store',
        type=_to_gb,
        help='upper bound memory limit for NiBabies processes',
    )
    g_perfm.add_argument(
        '--low-mem',
        action='store_true',
        help='attempt to reduce memory usage (will increase disk usage in working directory)',
    )
    g_perfm.add_argument(
        '--use-plugin',
        '--nipype-plugin-file',
        action='store',
        metavar='FILE',
        type=IsFile,
        help='nipype plugin configuration file',
    )
    g_perfm.add_argument(
        '--md-only-boilerplate',
        action='store_true',
        default=False,
        help='skip generation of HTML and LaTeX formatted citation with pandoc',
    )
    g_perfm.add_argument(
        '-v',
        '--verbose',
        dest='verbose_count',
        action='count',
        default=0,
        help='increases log verbosity for each occurrence, debug level is -vvv',
    )

    g_subset = parser.add_argument_group('Options for performing only a subset of the workflow')
    g_subset.add_argument('--anat-only', action='store_true', help='Run anatomical workflows only')
    g_subset.add_argument(
        '--level',
        action='store',
        default='full',
        choices=['minimal', 'resampling', 'full'],
        help="Processing level; may be 'minimal' (nothing that can be recomputed), "
        "'resampling' (recomputable targets that aid in resampling) "
        "or 'full' (all target outputs).",
    )
    g_subset.add_argument(
        '--boilerplate-only',
        '--boilerplate_only',
        action='store_true',
        default=False,
        help='Generate boilerplate only',
    )
    g_subset.add_argument(
        '--reports-only',
        action='store_true',
        default=False,
        help="Only generate reports, don't run workflows. This will only rerun report "
        'aggregation, not reportlet generation for specific nodes.',
    )

    g_conf = parser.add_argument_group('Workflow configuration')
    g_conf.add_argument(
        '--ignore',
        required=False,
        action='store',
        nargs='+',
        default=[],
        choices=['fieldmaps', 'slicetiming', 'sbref', 't1w', 't2w', 'flair', 'fmap-jacobian'],
        help='ignore selected aspects of the input dataset to disable corresponding '
        'parts of the workflow (a space delimited list)',
    )
    g_conf.add_argument(
        '--longitudinal',
        action='store_true',
        help='treat dataset as longitudinal - may increase runtime',
    )
    g_conf.add_argument(
        '--output-spaces',
        nargs='*',
        action=OutputReferencesAction,
        help="""\
Standard and non-standard spaces to resample anatomical and functional images to. \
Standard spaces may be specified by the form \
``<SPACE>[:cohort-<label>][:res-<resolution>][...]``, where ``<SPACE>`` is \
a keyword designating a spatial reference, and may be followed by optional, \
colon-separated parameters. \
Non-standard spaces imply specific orientations and sampling grids. \
Important to note, the ``res-*`` modifier does not define the resolution used for \
the spatial normalization. To generate no BOLD outputs, use this option without specifying \
any spatial references. For further details, please check out \
https://fmriprep.readthedocs.io/en/latest/spaces.html""",
    )
    g_conf.add_argument(
        '--me-output-echos',
        action='store_true',
        default=False,
        help="""\
Output individual echo time series with slice, motion and susceptibility correction. \
Useful for further Tedana processing post-NiBabies.""",
    )

    g_conf.add_argument(
        '--bold2anat-init',
        choices=['auto', 't1w', 't2w', 'header'],
        default='auto',
        help='Method of initial BOLD to anatomical coregistration. If `auto`, a T2w image is used '
        'if available, otherwise the T1w image. `t1w` forces use of the T1w, `t2w` forces use of '
        'the T2w, and `header` uses the BOLD header information without an initial registration.',
    )
    g_conf.add_argument(
        '--bold2anat-dof',
        action='store',
        default=6,
        choices=[6, 9, 12],
        type=int,
        help='Degrees of freedom when registering BOLD to anatomical images. '
        '6 degrees (rotation and translation) are used by default.',
    )
    g_conf.add_argument(
        '--force-bbr',
        action='store_true',
        dest='use_bbr',
        default=None,
        help='Always use boundary-based registration (no goodness-of-fit checks)',
    )
    g_conf.add_argument(
        '--force-no-bbr',
        action='store_false',
        dest='use_bbr',
        default=None,
        help='Do not use boundary-based registration (no goodness-of-fit checks)',
    )
    g_conf.add_argument(
        '--medial-surface-nan',
        required=False,
        action='store_true',
        default=False,
        help='Replace medial wall values with NaNs on functional GIFTI files. Only '
        'performed for GIFTI files mapped to a freesurfer subject (fsaverage or fsnative).',
    )
    g_conf.add_argument(
        '--project-goodvoxels',
        required=False,
        action='store_true',
        default=False,
        help='Exclude voxels whose timeseries have locally high coefficient of variation '
        'from surface resampling. Only performed for GIFTI files mapped to a freesurfer subject '
        '(fsaverage or fsnative).',
    )
    g_conf.add_argument(
        '--slice-time-ref',
        required=False,
        action='store',
        default=None,
        type=SliceTimeRef,
        help='The time of the reference slice to correct BOLD values to, as a fraction '
        'acquisition time. 0 indicates the start, 0.5 the midpoint, and 1 the end '
        'of acquisition. The alias `start` corresponds to 0, and `middle` to 0.5. '
        'The default value is 0.5.',
    )
    g_conf.add_argument(
        '--dummy-scans',
        required=False,
        action='store',
        default=None,
        type=int,
        help='Number of non steady state volumes.',
    )
    g_conf.add_argument(
        '--random-seed',
        dest='_random_seed',
        action='store',
        type=int,
        default=None,
        help='Initialize the random seed for the workflow',
    )
    g_conf.add_argument(
        '--me-t2s-fit-method',
        action='store',
        default='curvefit',
        choices=['curvefit', 'loglin'],
        help=(
            'The method by which to estimate T2* and S0 for multi-echo data. '
            "'curvefit' uses nonlinear regression. "
            "It is more memory intensive, but also may be more accurate, than 'loglin'. "
            "'loglin' uses log-linear regression. "
            'It is faster and less memory intensive, but may be less accurate.'
        ),
    )

    # Confounds options
    g_confounds = parser.add_argument_group('Specific options for estimating confounds')
    g_confounds.add_argument(
        '--return-all-components',
        dest='regressors_all_comps',
        required=False,
        action='store_true',
        default=False,
        help='Include all components estimated in CompCor decomposition in the confounds '
        'file instead of only the components sufficient to explain 50 percent of '
        'BOLD variance in each CompCor mask',
    )
    g_confounds.add_argument(
        '--fd-spike-threshold',
        dest='regressors_fd_th',
        required=False,
        action='store',
        default=0.5,
        type=float,
        help='Threshold for flagging a frame as an outlier on the basis of framewise displacement',
    )
    g_confounds.add_argument(
        '--dvars-spike-threshold',
        dest='regressors_dvars_th',
        required=False,
        action='store',
        default=1.5,
        type=float,
        help='Threshold for flagging a frame as an outlier on the basis of standardised DVARS',
    )

    #  ANTs options
    g_ants = parser.add_argument_group('Specific options for ANTs registrations')
    g_ants.add_argument(
        '--skull-strip-template',
        default='UNCInfant:cohort-1',
        type=Reference.from_string,
        help='select a template for skull-stripping with antsBrainExtraction',
    )
    g_ants.add_argument(
        '--skull-strip-fixed-seed',
        action='store_true',
        help='do not use a random seed for skull-stripping - will ensure '
        'run-to-run replicability when used with --omp-nthreads 1 and '
        'matching --random-seed <int>',
    )
    g_ants.add_argument(
        '--skull-strip-anat',
        action='store',
        choices=('auto', 'skip', 'force'),
        default='force',
        help="determiner for anatomical skull stripping ('force' ensures skull "
        "stripping, 'skip' ignores skull stripping, and 'auto' applies brain extraction "
        'based on the outcome of a heuristic to check whether the brain is already masked).',
    )

    # Fieldmap options
    g_fmap = parser.add_argument_group('Specific options for handling fieldmaps')
    g_fmap.add_argument(
        '--fmap-bspline',
        action='store_true',
        default=False,
        help='fit a B-Spline field using least-squares (experimental)',
    )
    g_fmap.add_argument(
        '--fmap-no-demean',
        action='store_false',
        default=True,
        help='do not remove median (within mask) from fieldmap',
    )

    # SyN-unwarp options
    g_syn = parser.add_argument_group('Specific options for SyN distortion correction')
    g_syn.add_argument(
        '--use-syn-sdc',
        nargs='?',
        choices=['warn', 'error'],
        action='store',
        const='error',
        default=False,
        help='Use fieldmap-less distortion correction based on anatomical image; '
        'if unable, error (default) or warn based on optional argument.',
    )
    g_syn.add_argument(
        '--force-syn',
        action='store_true',
        default=False,
        help='EXPERIMENTAL/TEMPORARY: Use SyN correction in addition to '
        'fieldmap correction, if available',
    )

    # FreeSurfer options
    g_fs = parser.add_argument_group('Specific options for FreeSurfer preprocessing')
    g_fs.add_argument(
        '--fs-license-file',
        metavar='FILE',
        type=IsFile,
        help='Path to FreeSurfer license key file. Get it (for free) by registering'
        ' at https://surfer.nmr.mgh.harvard.edu/registration.html',
    )
    g_fs.add_argument(
        '--fs-subjects-dir',
        metavar='PATH',
        type=Path,
        help='Path to existing FreeSurfer subjects directory to reuse. '
        '(default: OUTPUT_DIR/freesurfer)',
    )

    # Surface generation xor
    g_surfs = parser.add_argument_group('Surface preprocessing options')
    g_surfs.add_argument(
        '--no-submm-recon',
        action='store_false',
        dest='hires',
        help='disable sub-millimeter (hires) reconstruction',
    )
    g_surfs_xor = g_surfs.add_mutually_exclusive_group()
    g_surfs_xor.add_argument(
        '--cifti-output',
        nargs='?',
        const='91k',
        default=False,
        choices=('91k', '170k'),
        type=str,
        help='output preprocessed BOLD as a CIFTI dense timeseries. '
        'Optionally, the number of grayordinate can be specified '
        '(default is 91k, which equates to 2mm resolution)',
    )

    g_other = parser.add_argument_group('Other options')
    g_other.add_argument(
        '--output-layout',
        action='store',
        default='bids',
        choices=('bids', 'legacy'),
        help='Organization of outputs. bids (default) places NiBabies derivatives '
        'directly in the output directory, and defaults to placing FreeSurfer '
        'derivatives in <output-dir>/sourcedata/freesurfer. legacy creates derivative '
        'datasets as subdirectories of outputs.',
    )
    g_other.add_argument(
        '-w',
        '--work-dir',
        action='store',
        type=Path,
        default=Path('work').absolute(),
        help='path where intermediate results should be stored',
    )
    g_other.add_argument(
        '--clean-workdir',
        action=DeprecatedAction,
        default=False,
        help='Deprecated and has no effect.',
    )
    g_other.add_argument(
        '--resource-monitor',
        action='store_true',
        default=False,
        help="enable Nipype's resource monitoring to keep track of memory and CPU usage",
    )
    g_other.add_argument(
        '--config-file',
        action='store',
        metavar='FILE',
        help='Use pre-generated configuration file. Values in file will be overridden '
        'by command-line arguments.',
    )
    g_other.add_argument(
        '--write-graph',
        action='store_true',
        default=False,
        help='Write workflow graph.',
    )
    g_other.add_argument(
        '--stop-on-first-crash',
        action='store_true',
        default=False,
        help='Force stopping on first crash, even if a work directory was specified.',
    )
    g_other.add_argument(
        '--notrack',
        action='store_true',
        default=False,
        help='Opt-out of sending tracking information of this run to '
        'the NiBabies developers. This information helps to '
        'improve NiBabies and provides an indicator of real '
        'world usage crucial for obtaining funding.',
    )
    g_other.add_argument(
        '--debug',
        action='store',
        nargs='+',
        choices=config.DEBUG_MODES + ('all',),
        help="Debug mode(s) to enable. 'all' is alias for all available modes.",
    )

    g_other.add_argument(
        '--sloppy',
        action='store_true',
        default=False,
        help='Use low-quality tools for speed - TESTING ONLY',
    )

    latest = check_latest()
    if latest is not None and currentv < latest:
        print(
            f"""\
You are using NiBabies-{currentv}, and a newer version of NiBabies is available: {latest}.
Please check out our documentation about how and when to upgrade:
https://fmriprep.readthedocs.io/en/latest/faq.html#upgrading""",
            file=sys.stderr,
        )

    _blist = is_flagged()
    if _blist[0]:
        _reason = _blist[1] or 'unknown'
        print(
            f"""\
WARNING: Version {config.environment.version} of NiBabies (current) has been FLAGGED
(reason: {_reason}).
That means some severe flaw was found in it and we strongly
discourage its usage.""",
            file=sys.stderr,
        )

    # Add new options
    g_baby = parser.add_argument_group('NiBabies specific options')
    g_baby.add_argument(
        '--age-months',
        dest='age_months',
        type=int,
        help='Age in months',
    )
    g_baby.add_argument(
        '--segmentation-atlases-dir',
        type=DirNotEmpty,
        help='Directory containing precalculated segmentations to use for JointLabelFusion.',
    )
    g_baby.add_argument(
        '--fd-radius',
        type=float,
        default=45,
        help='Head radius in mm for framewise displacement calculation.',
    )
    g_baby.add_argument(
        '--deriv-filter-file',
        dest='derivatives_filters',
        type=_bids_filter,
        metavar='FILE',
        help='A JSON file for customizing the derivatives queries.',
    )
    g_baby.add_argument(
        '--surface-recon-method',
        choices=('auto', 'infantfs', 'freesurfer', 'mcribs', None),
        type=_str_none,
        default='auto',
        help='Method to use for surface reconstruction.',
    )
    g_baby.add_argument(
        '--reference-anat',
        '--reference-anatomical',
        choices=('T1w', 'T2w'),
        default=None,
        help='Override which anatomical to use as the structural reference. '
        'Generally, this is determined based on availability, age, and surface '
        'reconstruction method.',
    )
    g_baby.add_argument(
        '--hmc-bold-frame',
        type=int,
        default=16,
        help='Frame to start head motion estimation on BOLD.',
    )
    g_baby.add_argument(
        '--norm-csf',
        action='store_true',
        help='Replace low intensity voxels in CSF mask with average',
    )
    g_baby.add_argument(
        '--multi-step-reg',
        action=BooleanOptionalAction,
        default=True,
        help='For certain adult templates (MNI152NLin6Asym), perform two step '
        'registrations (native -> MNIInfant -> template) and concatenate into a single xfm',
    )
    return parser


def parse_args(args=None, namespace=None):
    """Parse args and run further checks on the command line."""
    import logging

    parser = _build_parser()
    opts = parser.parse_args(args, namespace)

    if opts.sloppy:
        config.loggers.cli.warning('Sloppy mode enabled: expect low-quality results')
        # disable multi-step registration as it expects a warp
        opts.multi_step_reg = False

    if opts.config_file:
        skip = {} if opts.reports_only else {'execution': ('run_uuid',)}
        config.load(opts.config_file, skip=skip)
        config.loggers.cli.info(f'Loaded previous configuration file {opts.config_file}')

    config.execution.log_level = int(max(25 - 5 * opts.verbose_count, logging.DEBUG))
    config.from_dict(vars(opts))

    # Retrieve logging level
    build_log = config.loggers.cli

    # Load base plugin_settings from file if --use-plugin
    if opts.use_plugin is not None:
        import yaml

        with open(opts.use_plugin) as f:
            plugin_settings = yaml.safe_load(f)
        _plugin = plugin_settings.get('plugin')
        if _plugin:
            config.nipype.plugin = _plugin
            config.nipype.plugin_args = plugin_settings.get('plugin_args', {})
            config.nipype.nprocs = opts.nprocs or config.nipype.plugin_args.get(
                'n_procs', config.nipype.nprocs
            )

    # Resource management options
    # Note that we're making strong assumptions about valid plugin args
    # This may need to be revisited if people try to use batch plugins
    if 1 < config.nipype.nprocs < config.nipype.omp_nthreads:
        build_log.warning(
            f'Per-process threads (--omp-nthreads={config.nipype.omp_nthreads}) exceed '
            f'total threads (--nthreads/--n_cpus={config.nipype.nprocs})'
        )

    # Inform the user about the risk of using brain-extracted images
    if config.workflow.skull_strip_anat == 'auto':
        build_log.warning(
            """\
Option ``--skull-strip-t1w`` was set to 'auto'. A heuristic will be \
applied to determine whether the input anatomical image(s) have already been skull-stripped.
If that were the case, brain extraction and INU correction will be skipped for those anatomical \
inputs. Please, BEWARE OF THE RISKS TO THE CONSISTENCY of results when using varying \
processing workflows across participants. To determine whether a participant has been run \
through the shortcut pipeline (meaning, brain extraction was skipped), please check the \
citation boilerplate. When reporting results with varying pipelines, please make sure you \
mention this particular variant of NiBabies listing the participants for which it was \
applied."""
        )

    bids_dir = config.execution.bids_dir
    output_dir = config.execution.output_dir
    work_dir = config.execution.work_dir
    version = config.environment.version
    output_layout = config.execution.output_layout

    if config.execution.fs_subjects_dir is None:
        if output_layout == 'bids':
            config.execution.fs_subjects_dir = output_dir / 'sourcedata' / 'freesurfer'
        elif output_layout == 'legacy':
            config.execution.fs_subjects_dir = output_dir / 'freesurfer'
    if config.execution.nibabies_dir is None:
        if output_layout == 'bids':
            config.execution.nibabies_dir = output_dir
        elif output_layout == 'legacy':
            config.execution.nibabies_dir = output_dir / 'nibabies'
    if config.workflow.surface_recon_method == 'mcribs':
        if output_layout == 'bids':
            config.execution.mcribs_dir = output_dir / 'sourcedata' / 'mcribs'
        elif output_layout == 'legacy':
            config.execution.mcribs_dir = output_dir / 'mcribs'
        # Ensure the directory is created
        config.execution.mcribs_dir.mkdir(exist_ok=True, parents=True)

    # Ensure input and output folders are not the same
    if output_dir == bids_dir:
        parser.error(
            'The selected output folder is the same as the input BIDS folder. '
            f'Please modify the output path (suggestion: {bids_dir}).'
            / 'derivatives'
            / f'nibabies-{version.split("+")[0]}'
        )

    if bids_dir in work_dir.parents:
        parser.error(
            'The selected working directory is a subdirectory of the input BIDS folder. '
            'Please modify the output path.'
        )

    # Validate inputs
    if not opts.skip_bids_validation:
        from ..utils.bids import validate_input_dir

        build_log.info(
            'Making sure the input data is BIDS compliant (warnings can be ignored in most cases).'
        )
        validate_input_dir(config.environment.exec_env, opts.bids_dir, opts.participant_label)

    # Setup directories
    config.execution.log_dir = config.execution.nibabies_dir / 'logs'
    # Check and create output and working directories
    config.execution.log_dir.mkdir(exist_ok=True, parents=True)
    work_dir.mkdir(exist_ok=True, parents=True)

    # Force initialization of the BIDSLayout
    config.execution.init()
    all_subjects = config.execution.layout.get_subjects(scope='raw')
    if config.execution.participant_label is None:
        config.execution.participant_label = all_subjects

    participant_label = set(config.execution.participant_label)
    missing_subjects = participant_label - set(all_subjects)
    if missing_subjects:
        parser.error(
            'One or more participant labels were not found in the BIDS directory: '
            f'{", ".join(missing_subjects)}.'
        )

    config.execution.participant_label = sorted(participant_label)

    config.execution.unique_labels = compute_subworkflows(
        layout=config.execution.layout,
        participant_ids=config.execution.participant_label,
        session_ids=config.execution.session_id,
    )
    config.workflow.skull_strip_template = config.workflow.skull_strip_template[0]

    # finally, write config to file
    config_file = config.execution.work_dir / config.execution.run_uuid / 'config.toml'
    config_file.parent.mkdir(exist_ok=True, parents=True)
    config.to_filename(config_file)


def compute_subworkflows(
    *,
    layout: BIDSLayout,
    participant_ids: list,
    session_ids: list | None = None,
) -> list:
    """
    Query all available participants and sessions, and construct the combinations of the
    subworkflows needed.
    """
    from niworkflows.utils.bids import collect_participants

    # consists of (subject_id, session_id) tuples
    subworkflows = []

    subject_list = collect_participants(layout, participant_ids, strict=True)
    for subject in subject_list:
        # Due to rapidly changing morphometry of the population
        # Ensure each subject session is processed individually
        sessions = session_ids or layout.get_sessions(scope='raw', subject=subject) or [None]
        for session in sessions:
            subworkflows.append((subject, session))
    return subworkflows



from nibabies.interfaces.bids import BIDSURI


def test_BIDSURI():
    """Test the BIDSURI interface."""

    dataset_links = {
        'raw': '/data',
        'deriv-0': '/data/derivatives/source-1',
    }
    out_dir = '/data/derivatives/nibabies'

    # A single element as a string
    interface = BIDSURI(
        numinputs=1,
        dataset_links=dataset_links,
        out_dir=out_dir,
    )
    interface.inputs.in1 = '/data/sub-01/func/sub-01_task-rest_bold.nii.gz'
    results = interface.run()
    assert results.outputs.out == ['bids:raw:sub-01/func/sub-01_task-rest_bold.nii.gz']

    # A single element as a list
    interface = BIDSURI(
        numinputs=1,
        dataset_links=dataset_links,
        out_dir=out_dir,
    )
    interface.inputs.in1 = ['/data/sub-01/func/sub-01_task-rest_bold.nii.gz']
    results = interface.run()
    assert results.outputs.out == ['bids:raw:sub-01/func/sub-01_task-rest_bold.nii.gz']

    # Two inputs: a string and a list
    interface = BIDSURI(
        numinputs=2,
        dataset_links=dataset_links,
        out_dir=out_dir,
    )
    interface.inputs.in1 = '/data/sub-01/func/sub-01_task-rest_bold.nii.gz'
    interface.inputs.in2 = [
        '/data/derivatives/source-1/sub-01/func/sub-01_task-rest_bold.nii.gz',
        '/out/sub-01/func/sub-01_task-rest_bold.nii.gz',
    ]
    results = interface.run()
    assert results.outputs.out == [
        'bids:raw:sub-01/func/sub-01_task-rest_bold.nii.gz',
        'bids:deriv-0:sub-01/func/sub-01_task-rest_bold.nii.gz',
        '/out/sub-01/func/sub-01_task-rest_bold.nii.gz',  # No change
    ]

    # Two inputs as lists
    interface = BIDSURI(
        numinputs=2,
        dataset_links=dataset_links,
        out_dir=out_dir,
    )
    interface.inputs.in1 = [
        '/data/sub-01/func/sub-01_task-rest_bold.nii.gz',
        'bids:raw:sub-01/func/sub-01_task-rest_boldref.nii.gz',
    ]
    interface.inputs.in2 = [
        '/data/derivatives/source-1/sub-01/func/sub-01_task-rest_bold.nii.gz',
        '/out/sub-01/func/sub-01_task-rest_bold.nii.gz',
    ]
    results = interface.run()
    assert results.outputs.out == [
        'bids:raw:sub-01/func/sub-01_task-rest_bold.nii.gz',
        'bids:raw:sub-01/func/sub-01_task-rest_boldref.nii.gz',  # No change
        'bids:deriv-0:sub-01/func/sub-01_task-rest_bold.nii.gz',
        '/out/sub-01/func/sub-01_task-rest_bold.nii.gz',  # No change
    ]



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""
Handling confounds.

    .. testsetup::

    >>> import os
    >>> import pandas as pd

"""

import os
import re

import nibabel as nb
import pandas as pd
from nipype import logging
from nipype.interfaces.base import (
    BaseInterfaceInputSpec,
    File,
    InputMultiObject,
    OutputMultiObject,
    SimpleInterface,
    TraitedSpec,
    isdefined,
    traits,
)
from nipype.utils.filemanip import fname_presuffix
from nireports.reportlets.modality.func import fMRIPlot
from niworkflows.utils.timeseries import _cifti_timeseries

LOGGER = logging.getLogger('nipype.interface')


class _aCompCorMasksInputSpec(BaseInterfaceInputSpec):
    in_vfs = InputMultiObject(File(exists=True), desc='Input volume fractions.')
    is_aseg = traits.Bool(
        False, usedefault=True, desc="Whether the input volume fractions come from FS' aseg."
    )
    bold_zooms = traits.Tuple(
        traits.Float, traits.Float, traits.Float, mandatory=True, desc='BOLD series zooms'
    )


class _aCompCorMasksOutputSpec(TraitedSpec):
    out_masks = OutputMultiObject(
        File(exists=True), desc='CSF, WM and combined masks, respectively'
    )


class aCompCorMasks(SimpleInterface):
    """Generate masks in T1w space for aCompCor."""

    input_spec = _aCompCorMasksInputSpec
    output_spec = _aCompCorMasksOutputSpec

    def _run_interface(self, runtime):
        from ..utils.confounds import acompcor_masks

        self._results['out_masks'] = acompcor_masks(
            self.inputs.in_vfs,
            self.inputs.is_aseg,
            self.inputs.bold_zooms,
        )
        return runtime


class _FilterDroppedInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, desc='input CompCor metadata')


class _FilterDroppedOutputSpec(TraitedSpec):
    out_file = File(desc='filtered CompCor metadata')


class FilterDropped(SimpleInterface):
    """Filter dropped components from CompCor metadata files

    Uses the boolean ``retained`` column to identify rows to keep or filter.
    """

    input_spec = _FilterDroppedInputSpec
    output_spec = _FilterDroppedOutputSpec

    def _run_interface(self, runtime):
        self._results['out_file'] = fname_presuffix(
            self.inputs.in_file, suffix='_filtered', use_ext=True, newpath=runtime.cwd
        )

        metadata = pd.read_csv(self.inputs.in_file, sep='\t')
        metadata[metadata.retained].to_csv(self._results['out_file'], sep='\t', index=False)

        return runtime


class _RenameACompCorInputSpec(BaseInterfaceInputSpec):
    components_file = File(exists=True, desc='input aCompCor components')
    metadata_file = File(exists=True, desc='input aCompCor metadata')


class _RenameACompCorOutputSpec(TraitedSpec):
    components_file = File(desc='output aCompCor components')
    metadata_file = File(desc='output aCompCor metadata')


class RenameACompCor(SimpleInterface):
    """Rename ACompCor components based on their masks

    Components from the "CSF" mask are ``c_comp_cor_*``.
    Components from the "WM" mask are ``w_comp_cor_*``.
    Components from the "combined" mask are ``a_comp_cor_*``.

    Each set of components is renumbered to start at ``?_comp_cor_00``.
    """

    input_spec = _RenameACompCorInputSpec
    output_spec = _RenameACompCorOutputSpec

    def _run_interface(self, runtime):
        try:
            components = pd.read_csv(self.inputs.components_file, sep='\t')
            metadata = pd.read_csv(self.inputs.metadata_file, sep='\t')
        except pd.errors.EmptyDataError:
            # Can occur when testing on short datasets; otherwise rare
            self._results['components_file'] = self.inputs.components_file
            self._results['metadata_file'] = self.inputs.metadata_file
            return runtime

        self._results['components_file'] = fname_presuffix(
            self.inputs.components_file, suffix='_renamed', use_ext=True, newpath=runtime.cwd
        )
        self._results['metadata_file'] = fname_presuffix(
            self.inputs.metadata_file, suffix='_renamed', use_ext=True, newpath=runtime.cwd
        )

        all_comp_cor = metadata[metadata['retained']]

        c_comp_cor = all_comp_cor[all_comp_cor['mask'] == 'CSF']
        w_comp_cor = all_comp_cor[all_comp_cor['mask'] == 'WM']
        a_comp_cor = all_comp_cor[all_comp_cor['mask'] == 'combined']

        c_orig = c_comp_cor['component']
        c_new = [f'c_comp_cor_{i:02d}' for i in range(len(c_orig))]

        w_orig = w_comp_cor['component']
        w_new = [f'w_comp_cor_{i:02d}' for i in range(len(w_orig))]

        a_orig = a_comp_cor['component']
        a_new = [f'a_comp_cor_{i:02d}' for i in range(len(a_orig))]

        final_components = components.rename(columns=dict(zip(c_orig, c_new, strict=False)))
        final_components.rename(columns=dict(zip(w_orig, w_new, strict=False)), inplace=True)
        final_components.rename(columns=dict(zip(a_orig, a_new, strict=False)), inplace=True)
        final_components.to_csv(self._results['components_file'], sep='\t', index=False)

        metadata.loc[c_comp_cor.index, 'component'] = c_new
        metadata.loc[w_comp_cor.index, 'component'] = w_new
        metadata.loc[a_comp_cor.index, 'component'] = a_new

        metadata.to_csv(self._results['metadata_file'], sep='\t', index=False)

        return runtime


class GatherConfoundsInputSpec(BaseInterfaceInputSpec):
    signals = File(exists=True, desc='input signals')
    dvars = File(exists=True, desc='file containing DVARS')
    std_dvars = File(exists=True, desc='file containing standardized DVARS')
    fd = File(exists=True, desc='input framewise displacement')
    rmsd = File(exists=True, desc='input RMS framewise displacement')
    tcompcor = File(exists=True, desc='input tCompCorr')
    acompcor = File(exists=True, desc='input aCompCorr')
    crowncompcor = File(exists=True, desc='input crown-based regressors')
    cos_basis = File(exists=True, desc='input cosine basis')
    motion = File(exists=True, desc='input motion parameters')


class GatherConfoundsOutputSpec(TraitedSpec):
    confounds_file = File(exists=True, desc='output confounds file')
    confounds_list = traits.List(traits.Str, desc='list of headers')


class GatherConfounds(SimpleInterface):
    r"""
    Combine various sources of confounds in one TSV file

    >>> pd.DataFrame({'a': [0.1]}).to_csv('signals.tsv', index=False, na_rep='n/a')
    >>> pd.DataFrame({'b': [0.2]}).to_csv('dvars.tsv', index=False, na_rep='n/a')

    >>> gather = GatherConfounds()
    >>> gather.inputs.signals = 'signals.tsv'
    >>> gather.inputs.dvars = 'dvars.tsv'
    >>> res = gather.run()
    >>> res.outputs.confounds_list
    ['Global signals', 'DVARS']

    >>> pd.read_csv(res.outputs.confounds_file, sep='\s+', index_col=None,
    ...             engine='python')  # doctest: +NORMALIZE_WHITESPACE
         a    b
    0  0.1  0.2

    """

    input_spec = GatherConfoundsInputSpec
    output_spec = GatherConfoundsOutputSpec

    def _run_interface(self, runtime):
        combined_out, confounds_list = _gather_confounds(
            signals=self.inputs.signals,
            dvars=self.inputs.dvars,
            std_dvars=self.inputs.std_dvars,
            fdisp=self.inputs.fd,
            rmsd=self.inputs.rmsd,
            tcompcor=self.inputs.tcompcor,
            acompcor=self.inputs.acompcor,
            crowncompcor=self.inputs.crowncompcor,
            cos_basis=self.inputs.cos_basis,
            motion=self.inputs.motion,
            newpath=runtime.cwd,
        )
        self._results['confounds_file'] = combined_out
        self._results['confounds_list'] = confounds_list
        return runtime


def _gather_confounds(
    signals=None,
    dvars=None,
    std_dvars=None,
    fdisp=None,
    rmsd=None,
    tcompcor=None,
    acompcor=None,
    crowncompcor=None,
    cos_basis=None,
    motion=None,
    newpath=None,
):
    r"""
    Load confounds from the filenames, concatenate together horizontally
    and save new file.

    >>> pd.DataFrame({'Global Signal': [0.1]}).to_csv('signals.tsv', index=False, na_rep='n/a')
    >>> pd.DataFrame({'stdDVARS': [0.2]}).to_csv('dvars.tsv', index=False, na_rep='n/a')
    >>> out_file, confound_list = _gather_confounds('signals.tsv', 'dvars.tsv')
    >>> confound_list
    ['Global signals', 'DVARS']

    >>> pd.read_csv(out_file, sep='\s+', index_col=None,
    ...             engine='python')  # doctest: +NORMALIZE_WHITESPACE
       global_signal  std_dvars
    0            0.1        0.2


    """

    def less_breakable(a_string):
        """hardens the string to different envs (i.e., case insensitive, no whitespace, '#'"""
        return ''.join(a_string.split()).strip('#')

    # Taken from https://stackoverflow.com/questions/1175208/
    # If we end up using it more than just here, probably worth pulling in a well-tested package
    def camel_to_snake(name):
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

    def _adjust_indices(left_df, right_df):
        # This forces missing values to appear at the beginning of the DataFrame
        # instead of the end
        index_diff = len(left_df.index) - len(right_df.index)
        if index_diff > 0:
            right_df.index = range(index_diff, len(right_df.index) + index_diff)
        elif index_diff < 0:
            left_df.index = range(-index_diff, len(left_df.index) - index_diff)

    all_files = []
    confounds_list = []
    for confound, name in (
        (signals, 'Global signals'),
        (std_dvars, 'Standardized DVARS'),
        (dvars, 'DVARS'),
        (fdisp, 'Framewise displacement'),
        (rmsd, 'Framewise displacement (RMS)'),
        (tcompcor, 'tCompCor'),
        (acompcor, 'aCompCor'),
        (crowncompcor, 'crownCompCor'),
        (cos_basis, 'Cosine basis'),
        (motion, 'Motion parameters'),
    ):
        if confound is not None and isdefined(confound):
            confounds_list.append(name)
            if os.path.exists(confound) and os.stat(confound).st_size > 0:
                all_files.append(confound)

    confounds_data = pd.DataFrame()
    for file_name in all_files:  # assumes they all have headings already
        try:
            new = pd.read_csv(file_name, sep='\t')
        except pd.errors.EmptyDataError:
            # No data, nothing to concat
            continue
        for column_name in new.columns:
            new.rename(
                columns={column_name: camel_to_snake(less_breakable(column_name))}, inplace=True
            )

        _adjust_indices(confounds_data, new)
        confounds_data = pd.concat((confounds_data, new), axis=1)

    if newpath is None:
        newpath = os.getcwd()

    combined_out = os.path.join(newpath, 'confounds.tsv')
    confounds_data.to_csv(combined_out, sep='\t', index=False, na_rep='n/a')

    return combined_out, confounds_list


class _FMRISummaryInputSpec(BaseInterfaceInputSpec):
    in_nifti = File(exists=True, desc='input BOLD (4D NIfTI file)')
    in_cifti = File(exists=True, mandatory=True, desc='input BOLD (CIFTI dense timeseries)')
    in_segm = File(exists=True, desc='volumetric segmentation corresponding to in_nifti')
    confounds_file = File(exists=True, mandatory=True, desc="BIDS' _confounds.tsv file")

    str_or_tuple = traits.Either(
        traits.Str,
        traits.Tuple(traits.Str, traits.Either(None, traits.Str)),
        traits.Tuple(traits.Str, traits.Either(None, traits.Str), traits.Either(None, traits.Str)),
    )
    confounds_list = traits.List(
        str_or_tuple, minlen=1, desc='list of headers to extract from the confounds_file'
    )
    tr = traits.Either(None, traits.Float, usedefault=True, desc='the repetition time')
    drop_trs = traits.Int(0, usedefault=True, desc='dummy scans')


class _FMRISummaryOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc='written file path')


class FMRISummary(SimpleInterface):
    """
    Copy the x-form matrices from `hdr_file` to `out_file`.
    """

    input_spec = _FMRISummaryInputSpec
    output_spec = _FMRISummaryOutputSpec

    def _run_interface(self, runtime):
        self._results['out_file'] = fname_presuffix(
            self.inputs.in_cifti, suffix='_fmriplot.svg', use_ext=False, newpath=runtime.cwd
        )

        # Read input object and create timeseries + segments object
        # seg_file = self.inputs.in_segm or None

        # MG: Only plot CIFTI outputs for the time being
        cifti_data, cifti_segments = _cifti_timeseries(nb.load(self.inputs.in_cifti))

        # if seg_file is not None:
        #     # Append WM+CSF and Edge masks
        #     cifti_length = cifti_data.shape[0]
        #     dataset = np.vstack((cifti_data, dataset))
        #     segments = {k: np.array(v) + cifti_length for k, v in segments.items()}
        #     cifti_segments.update(segments)
        #     segments = cifti_segments
        # else:

        # MG: This is the only route atm
        dataset, segments = cifti_data, cifti_segments

        dataframe = pd.read_csv(
            self.inputs.confounds_file,
            sep='\t',
            index_col=None,
            dtype='float32',
            na_filter=True,
            na_values='n/a',
        )

        headers = []
        units = {}
        names = {}

        for conf_el in self.inputs.confounds_list:
            if isinstance(conf_el, list | tuple):
                headers.append(conf_el[0])
                if conf_el[1] is not None:
                    units[conf_el[0]] = conf_el[1]

                if len(conf_el) > 2 and conf_el[2] is not None:
                    names[conf_el[0]] = conf_el[2]
            else:
                headers.append(conf_el)

        if not headers:
            data = None
            units = None
        else:
            data = dataframe[headers]

        data = data.rename(columns=names)

        fig = fMRIPlot(
            dataset,
            segments=segments,
            tr=self.inputs.tr,
            confounds=data,
            units=units,
            nskip=self.inputs.drop_trs,
            paired_carpet=True,
        ).plot()
        fig.savefig(self._results['out_file'], bbox_inches='tight')
        return runtime



import typing as ty

import nipype.interfaces.utility as niu
import nipype.pipeline.engine as pe
from niworkflows.engine import tag
from smriprep.workflows.outputs import (
    init_ds_anat_volumes_wf,
    init_ds_grayord_metrics_wf,
    init_ds_surface_metrics_wf,
    init_ds_surfaces_wf,
)
from smriprep.workflows.surfaces import (
    init_hcp_morphometrics_wf,
    init_morph_grayords_wf,
    init_resample_surfaces_wf,
    init_surface_derivatives_wf,
)

from nibabies import config
from nibabies.workflows.anatomical.outputs import init_ds_seg_wf
from nibabies.workflows.anatomical.surfaces import init_resample_surfaces_dhcp_wf

if ty.TYPE_CHECKING:
    from niworkflows.utils.spaces import SpatialReferences

LOGGER = config.loggers.workflow


@tag('anat.apply')
def init_infant_anat_apply_wf(
    *,
    bids_root: str,
    cifti_output: ty.Literal['91k', '170k', False],
    msm_sulc: bool,
    omp_nthreads: int,
    output_dir: str,
    recon_method: ty.Literal['freesurfer', 'infantfs', 'mcribs', None],
    reference_anat: ty.Literal['T1w', 'T2w'],
    precomputed: dict,
    spaces: 'SpatialReferences',
    name: str = 'infant_anat_apply_wf',
) -> pe.Workflow:
    """Save superfluous outputs."""
    workflow = pe.Workflow(name=name)

    reg_sphere = f'sphere_reg_{"msm" if msm_sulc else "fsLR"}'
    inputnode = pe.Node(
        niu.IdentityInterface(
            fields=[
                'anat2std_xfm',
                'anat_valid_list',
                'anat_preproc',
                'anat_mask',
                'anat_dseg',
                'anat_tpms',
                'subjects_dir',
                'subject_id',
                'fsnative2anat_xfm',
                'sulc',
                'template',
                'thickness',
                'white',
                'pial',
                'midthickness',
                reg_sphere,
                # template workflow inputs
                'std_t1w',
                'anat2std_xfm',
                'std_space',
                'std_cohort',
                'std_resolution',
            ]
        ),
        name='inputnode',
    )

    outputnode = pe.Node(
        niu.IdentityInterface(fields=['anat_aseg', 'anat_aparc', 'midthickness_fsLR', 'roi']),
        name='outputnode',
    )

    if spaces.cached.get_spaces(nonstandard=False, dim=(3,)):
        ds_std_volumes_wf = init_ds_anat_volumes_wf(
            bids_root=bids_root,
            output_dir=output_dir,
            name='ds_std_volumes_wf',
        )

        workflow.connect([
            (inputnode, ds_std_volumes_wf, [
                ('anat_valid_list', 'inputnode.source_files'),
                ('anat_preproc', 'inputnode.anat_preproc'),
                ('anat_mask', 'inputnode.anat_mask'),
                ('anat_dseg', 'inputnode.anat_dseg'),
                ('anat_tpms', 'inputnode.anat_tpms'),
            ]),
            (inputnode, ds_std_volumes_wf, [
                ('std_t1w', 'inputnode.ref_file'),
                ('anat2std_xfm', 'inputnode.anat2std_xfm'),
                ('std_space', 'inputnode.space'),
                ('std_cohort', 'inputnode.cohort'),
                ('std_resolution', 'inputnode.resolution'),
            ]),
        ])  # fmt:skip

    if recon_method is not None:
        anat_aseg = precomputed.get(f'{reference_anat.lower()}_aseg')
        anat_aparc = precomputed.get(f'{reference_anat.lower()}_aparc')

        surface_derivatives_wf = init_surface_derivatives_wf()
        # Split out segmentations to better save precomputed inputs
        seg_buffer = pe.Node(
            niu.IdentityInterface(fields=['anat_aseg', 'anat_aparc']),
            name='seg_buffer',
        )

        ds_aseg_wf = init_ds_seg_wf(
            output_dir=output_dir,
            seg_type='aseg',
            extra_entities={'space': reference_anat},
        )
        ds_aparc_wf = init_ds_seg_wf(
            output_dir=output_dir,
            seg_type='aparcaseg',
            extra_entities={'space': reference_anat},
        )
        if anat_aseg:
            LOGGER.info('ANAT - using precomputed aseg')
            seg_buffer.inputs.anat_aseg = anat_aseg
        else:
            workflow.connect([
                (surface_derivatives_wf, seg_buffer, [
                    ('outputnode.out_aseg', 'anat_aseg'),
                ]),
            ])  # fmt:skip

        if anat_aparc:
            LOGGER.info('ANAT - using precomputed aparc')
            seg_buffer.inputs.anat_aparc = anat_aparc
        else:
            workflow.connect([
                (surface_derivatives_wf, seg_buffer, [
                    ('outputnode.out_aparc', 'anat_aparc'),
                ]),
            ])  # fmt:skip

        ds_surfaces_wf = init_ds_surfaces_wf(output_dir=output_dir, surfaces=['inflated'])
        ds_curv_wf = init_ds_surface_metrics_wf(
            bids_root=bids_root, output_dir=output_dir, metrics=['curv'], name='ds_curv_wf'
        )

        workflow.connect([
            (inputnode, surface_derivatives_wf, [
                ('anat_preproc', 'inputnode.reference'),
                ('subjects_dir', 'inputnode.subjects_dir'),
                ('subject_id', 'inputnode.subject_id'),
                ('fsnative2anat_xfm', 'inputnode.fsnative2anat_xfm'),
            ]),
            (inputnode, ds_surfaces_wf, [
                ('anat_valid_list', 'inputnode.source_files'),
            ]),
            (surface_derivatives_wf, ds_surfaces_wf, [
                ('outputnode.inflated', 'inputnode.inflated'),
            ]),
            (inputnode, ds_curv_wf, [
                ('anat_valid_list', 'inputnode.source_files'),
            ]),
            (surface_derivatives_wf, ds_curv_wf, [
                ('outputnode.curv', 'inputnode.curv'),
            ]),
            (inputnode, ds_aseg_wf, [
                ('anat_valid_list', 'inputnode.source_files'),
            ]),
            (inputnode, ds_aparc_wf, [
                ('anat_valid_list', 'inputnode.source_files'),
            ]),
            (seg_buffer, outputnode, [
                ('anat_aparc', 'anat_aparc'),
                ('anat_aseg', 'anat_aseg'),
            ]),
            (seg_buffer, ds_aparc_wf, [
                ('anat_aparc', 'inputnode.in_seg'),
            ]),
            (seg_buffer, ds_aseg_wf, [
                ('anat_aseg', 'inputnode.in_seg'),
            ])
        ])  # fmt:skip

        if cifti_output:
            hcp_morphometrics_wf = init_hcp_morphometrics_wf(omp_nthreads=omp_nthreads)
            if recon_method == 'mcribs':
                resample_surfaces_wf = init_resample_surfaces_dhcp_wf(
                    surfaces=['white', 'pial', 'midthickness'],
                    grayord_density=cifti_output,
                )
            else:
                resample_surfaces_wf = init_resample_surfaces_wf(
                    surfaces=['white', 'pial', 'midthickness'], grayord_density=cifti_output
                )
            morph_grayords_wf = init_morph_grayords_wf(
                grayord_density=cifti_output, omp_nthreads=omp_nthreads
            )

            ds_fsLR_surfaces_wf = init_ds_surfaces_wf(
                output_dir=output_dir,
                surfaces=['white', 'pial', 'midthickness'],
                entities={
                    'space': 'dhcpAsym' if recon_method == 'mcribs' else 'fsLR',
                    'density': '32k' if cifti_output == '91k' else '59k',
                },
                name='ds_fsLR_surfaces_wf',
            )

            ds_grayord_metrics_wf = init_ds_grayord_metrics_wf(
                bids_root=bids_root,
                output_dir=output_dir,
                metrics=['curv', 'thickness', 'sulc'],
                cifti_output=cifti_output,
            )

            workflow.connect([
                (inputnode, hcp_morphometrics_wf, [
                    ('subject_id', 'inputnode.subject_id'),
                    ('sulc', 'inputnode.sulc'),
                    ('thickness', 'inputnode.thickness'),
                    ('midthickness', 'inputnode.midthickness'),
                ]),
                (surface_derivatives_wf, hcp_morphometrics_wf, [
                    ('outputnode.curv', 'inputnode.curv'),
                ]),
                (inputnode, resample_surfaces_wf, [
                    ('white', 'inputnode.white'),
                    ('pial', 'inputnode.pial'),
                    ('midthickness', 'inputnode.midthickness'),
                    (reg_sphere, 'inputnode.sphere_reg_fsLR'),
                ]),
                (inputnode, morph_grayords_wf, [
                    ('midthickness', 'inputnode.midthickness'),
                    (reg_sphere, 'inputnode.sphere_reg_fsLR'),
                ]),
                (hcp_morphometrics_wf, morph_grayords_wf, [
                    ('outputnode.curv', 'inputnode.curv'),
                    ('outputnode.sulc', 'inputnode.sulc'),
                    ('outputnode.thickness', 'inputnode.thickness'),
                    ('outputnode.roi', 'inputnode.roi'),
                ]),
                (hcp_morphometrics_wf, outputnode, [
                    ('outputnode.roi', 'roi'),
                ]),
                (resample_surfaces_wf, morph_grayords_wf, [
                    ('outputnode.midthickness_fsLR', 'inputnode.midthickness_fsLR'),
                ]),
                (inputnode, ds_fsLR_surfaces_wf, [
                    ('anat_valid_list', 'inputnode.source_files'),
                ]),
                (resample_surfaces_wf, outputnode, [
                    ('outputnode.midthickness_fsLR', 'midthickness_fsLR'),
                ]),
                (resample_surfaces_wf, ds_fsLR_surfaces_wf, [
                    ('outputnode.white_fsLR', 'inputnode.white'),
                    ('outputnode.pial_fsLR', 'inputnode.pial'),
                    ('outputnode.midthickness_fsLR', 'inputnode.midthickness'),
                ]),
                (inputnode, ds_grayord_metrics_wf, [
                    ('anat_valid_list', 'inputnode.source_files'),
                ]),
                (morph_grayords_wf, ds_grayord_metrics_wf, [
                    ('outputnode.curv_fsLR', 'inputnode.curv'),
                    ('outputnode.curv_metadata', 'inputnode.curv_metadata'),
                    ('outputnode.thickness_fsLR', 'inputnode.thickness'),
                    ('outputnode.thickness_metadata', 'inputnode.thickness_metadata'),
                    ('outputnode.sulc_fsLR', 'inputnode.sulc'),
                    ('outputnode.sulc_metadata', 'inputnode.sulc_metadata'),
                ]),
            ])  # fmt:skip

    return workflow



"""Interfaces for resampling images in a single shot"""

import asyncio
import os
from collections.abc import Callable
from functools import partial
from typing import TypeVar

import nibabel as nb
import nitransforms as nt
import numpy as np
from nipype.interfaces.base import (
    File,
    InputMultiObject,
    SimpleInterface,
    TraitedSpec,
    traits,
)
from nipype.utils.filemanip import fname_presuffix
from scipy import ndimage as ndi
from scipy.sparse import hstack as sparse_hstack
from sdcflows.transform import grid_bspline_weights
from sdcflows.utils.tools import ensure_positive_cosines

from nibabies.utils.transforms import load_transforms

R = TypeVar('R')


async def worker(job: Callable[[], R], semaphore: asyncio.Semaphore) -> R:
    async with semaphore:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, job)


class ResampleSeriesInputSpec(TraitedSpec):
    in_file = File(exists=True, mandatory=True, desc='3D or 4D image file to resample')
    ref_file = File(exists=True, mandatory=True, desc='File to resample in_file to')
    transforms = InputMultiObject(
        File(exists=True),
        mandatory=True,
        desc='Transform files, from in_file to ref_file (image mode)',
    )
    inverse = InputMultiObject(
        traits.Bool,
        value=[False],
        usedefault=True,
        desc='Whether to invert each file in transforms',
    )
    fieldmap = File(exists=True, desc='Fieldmap file resampled into reference space')
    ro_time = traits.Float(desc='EPI readout time (s).')
    pe_dir = traits.Enum(
        'i',
        'i-',
        'j',
        'j-',
        'k',
        'k-',
        desc='the phase-encoding direction corresponding to in_data',
    )
    jacobian = traits.Bool(mandatory=True, desc='Whether to apply Jacobian correction')
    num_threads = traits.Int(1, usedefault=True, desc='Number of threads to use for resampling')
    output_data_type = traits.Str('float32', usedefault=True, desc='Data type of output image')
    order = traits.Int(3, usedefault=True, desc='Order of interpolation (0=nearest, 3=cubic)')
    mode = traits.Str(
        'constant',
        usedefault=True,
        desc='How data is extended beyond its boundaries. '
        'See scipy.ndimage.map_coordinates for more details.',
    )
    cval = traits.Float(0.0, usedefault=True, desc='Value to fill past edges of data')
    prefilter = traits.Bool(True, usedefault=True, desc='Spline-prefilter data if order > 1')


class ResampleSeriesOutputSpec(TraitedSpec):
    out_file = File(desc='Resampled image or series')


class ResampleSeries(SimpleInterface):
    """Resample a time series, applying susceptibility and motion correction
    simultaneously.
    """

    input_spec = ResampleSeriesInputSpec
    output_spec = ResampleSeriesOutputSpec

    def _run_interface(self, runtime):
        out_path = fname_presuffix(self.inputs.in_file, suffix='resampled', newpath=runtime.cwd)

        source = nb.load(self.inputs.in_file)
        target = nb.load(self.inputs.ref_file)
        fieldmap = nb.load(self.inputs.fieldmap) if self.inputs.fieldmap else None

        nvols = source.shape[3] if source.ndim > 3 else 1

        transforms = load_transforms(self.inputs.transforms, self.inputs.inverse)

        pe_dir = self.inputs.pe_dir
        ro_time = self.inputs.ro_time
        pe_info = None

        if pe_dir and ro_time:
            pe_axis = 'ijk'.index(pe_dir[0])
            pe_flip = pe_dir.endswith('-')

            # Nitransforms displacements are positive
            source, axcodes = ensure_positive_cosines(source)
            axis_flip = axcodes[pe_axis] in 'LPI'

            pe_info = [(pe_axis, -ro_time if (axis_flip ^ pe_flip) else ro_time)] * nvols

        resampled = resample_image(
            source=source,
            target=target,
            transforms=transforms,
            fieldmap=fieldmap,
            pe_info=pe_info,
            jacobian=self.inputs.jacobian,
            nthreads=self.inputs.num_threads,
            output_dtype=self.inputs.output_data_type,
            order=self.inputs.order,
            mode=self.inputs.mode,
            cval=self.inputs.cval,
            prefilter=self.inputs.prefilter,
        )
        resampled.to_filename(out_path)

        self._results['out_file'] = out_path
        return runtime


class ReconstructFieldmapInputSpec(TraitedSpec):
    in_coeffs = InputMultiObject(
        File(exists=True), mandatory=True, desc='SDCflows-style spline coefficient files'
    )
    target_ref_file = File(
        exists=True, mandatory=True, desc='Image to reconstruct the field in alignment with'
    )
    fmap_ref_file = File(
        exists=True, mandatory=True, desc='Reference file aligned with coefficients'
    )
    transforms = InputMultiObject(
        File(exists=True),
        mandatory=True,
        desc='Transform files, from in_file to ref_file (image mode)',
    )
    inverse = InputMultiObject(
        traits.Bool,
        value=[False],
        usedefault=True,
        desc='Whether to invert each file in transforms',
    )


class ReconstructFieldmapOutputSpec(TraitedSpec):
    out_file = File(desc='Fieldmap reconstructed in target_ref_file space')


class ReconstructFieldmap(SimpleInterface):
    """Reconstruct a fieldmap from B-spline coefficients in a target space.

    If the target reference does not have an aligned grid (guaranteed if
    transforms include a warp), then a reference file describing the space
    where it is valid to extrapolate the field will be used as an intermediate
    step.
    """

    input_spec = ReconstructFieldmapInputSpec
    output_spec = ReconstructFieldmapOutputSpec

    def _run_interface(self, runtime):
        out_path = fname_presuffix(self.inputs.in_coeffs[-1], suffix='rec', newpath=runtime.cwd)

        coefficients = [nb.load(coeff_file) for coeff_file in self.inputs.in_coeffs]
        target = nb.load(self.inputs.target_ref_file)
        fmapref = nb.load(self.inputs.fmap_ref_file)

        transforms = load_transforms(self.inputs.transforms, self.inputs.inverse)

        fieldmap = reconstruct_fieldmap(
            coefficients=coefficients,
            fmap_reference=fmapref,
            target=target,
            transforms=transforms,
        )
        fieldmap.to_filename(out_path)

        self._results['out_file'] = out_path
        return runtime


class DistortionParametersInputSpec(TraitedSpec):
    in_file = File(exists=True, desc='EPI image corresponding to the metadata')
    metadata = traits.Dict(mandatory=True, desc='metadata corresponding to the inputs')


class DistortionParametersOutputSpec(TraitedSpec):
    readout_time = traits.Float
    pe_direction = traits.Enum('i', 'i-', 'j', 'j-', 'k', 'k-')


class DistortionParameters(SimpleInterface):
    """Retrieve PhaseEncodingDirection and TotalReadoutTime from available metadata.

    One or both parameters may be missing; downstream interfaces should be prepared
    to handle this.
    """

    input_spec = DistortionParametersInputSpec
    output_spec = DistortionParametersOutputSpec

    def _run_interface(self, runtime):
        from sdcflows.utils.epimanip import get_trt

        try:
            self._results['readout_time'] = get_trt(
                self.inputs.metadata,
                self.inputs.in_file or None,
            )
            self._results['pe_direction'] = self.inputs.metadata['PhaseEncodingDirection']
        except (KeyError, ValueError):
            pass

        return runtime


def resample_vol(
    data: np.ndarray,
    coordinates: np.ndarray,
    pe_info: tuple[int, float],
    jacobian: bool,
    hmc_xfm: np.ndarray | None,
    fmap_hz: np.ndarray,
    output: np.dtype | np.ndarray | None = None,
    order: int = 3,
    mode: str = 'constant',
    cval: float = 0.0,
    prefilter: bool = True,
) -> np.ndarray:
    """Resample a volume at specified coordinates

    This function implements simultaneous head-motion correction and
    susceptibility-distortion correction. It accepts coordinates in
    the source voxel space. It is the responsibility of the caller to
    transform coordinates from any other target space.

    Parameters
    ----------
    data
        The data array to resample
    coordinates
        The first-approximation voxel coordinates to sample from ``data``
        The first dimension should have length ``data.ndim``. The further
        dimensions have the shape of the target array.
    pe_info
        The readout vector in the form of (axis, signed-readout-time)
        ``(1, -0.04)`` becomes ``[0, -0.04, 0]``, which indicates that a
        +1 Hz deflection in the field shifts 0.04 voxels toward the start
        of the data array in the second dimension.
    hmc_xfm
        Affine transformation accounting for head motion from the individual
        volume into the BOLD reference space. This affine must be in VOX2VOX
        form.
    fmap_hz
        The fieldmap, sampled to the target space, in Hz
    output
        The dtype or a pre-allocated array for sampling into the target space.
        If pre-allocated, ``output.shape == coordinates.shape[1:]``.
    order
        Order of interpolation (default: 3 = cubic)
    mode
        How ``data`` is extended beyond its boundaries. See
        :func:`scipy.ndimage.map_coordinates` for more details.
    cval
        Value to fill past edges of ``data`` if ``mode`` is ``'constant'``.
    prefilter
        Determines if ``data`` is pre-filtered before interpolation.

    Returns
    -------
    resampled_array
        The resampled array, with shape ``coordinates.shape[1:]``.
    """
    if hmc_xfm is not None:
        # Move image with the head
        coords_shape = coordinates.shape
        coordinates = nb.affines.apply_affine(
            hmc_xfm, coordinates.reshape(coords_shape[0], -1).T
        ).T.reshape(coords_shape)
    else:
        # Copy coordinates to avoid interfering with other calls
        coordinates = coordinates.copy()

    vsm = fmap_hz * pe_info[1]
    coordinates[pe_info[0], ...] += vsm

    result = ndi.map_coordinates(
        data,
        coordinates,
        output=output,
        order=order,
        mode=mode,
        cval=cval,
        prefilter=prefilter,
    )

    if jacobian:
        result *= 1 + np.gradient(vsm, axis=pe_info[0])

    return result


async def resample_series_async(
    data: np.ndarray,
    coordinates: np.ndarray,
    pe_info: list[tuple[int, float]],
    jacobian: bool,
    hmc_xfms: list[np.ndarray] | None,
    fmap_hz: np.ndarray,
    output_dtype: np.dtype | None = None,
    order: int = 3,
    mode: str = 'constant',
    cval: float = 0.0,
    prefilter: bool = True,
    max_concurrent: int = min(os.cpu_count(), 12),
) -> np.ndarray:
    """Resample a 4D time series at specified coordinates

    This function implements simultaneous head-motion correction and
    susceptibility-distortion correction. It accepts coordinates in
    the source voxel space. It is the responsibility of the caller to
    transform coordinates from any other target space.

    Parameters
    ----------
    data
        The data array to resample
    coordinates
        The first-approximation voxel coordinates to sample from ``data``.
        The first dimension should have length 3.
        The further dimensions determine the shape of the target array.
    pe_info
        A list of readout vectors in the form of (axis, signed-readout-time)
        ``(1, -0.04)`` becomes ``[0, -0.04, 0]``, which indicates that a
        +1 Hz deflection in the field shifts 0.04 voxels toward the start
        of the data array in the second dimension.
    hmc_xfm
        A sequence of affine transformations accounting for head motion from
        the individual volume into the BOLD reference space.
        These affines must be in VOX2VOX form.
    fmap_hz
        The fieldmap, sampled to the target space, in Hz
    output_dtype
        The dtype of the output array.
    order
        Order of interpolation (default: 3 = cubic)
    mode
        How ``data`` is extended beyond its boundaries. See
        :func:`scipy.ndimage.map_coordinates` for more details.
    cval
        Value to fill past edges of ``data`` if ``mode`` is ``'constant'``.
    prefilter
        Determines if ``data`` is pre-filtered before interpolation.
    max_concurrent
        Maximum number of volumes to resample concurrently

    Returns
    -------
    resampled_array
        The resampled array, with shape ``coordinates.shape[1:] + (N,)``,
        where N is the number of volumes in ``data``.
    """
    if data.ndim == 3:
        return resample_vol(
            data,
            coordinates,
            pe_info[0],
            jacobian,
            hmc_xfms[0] if hmc_xfms else None,
            fmap_hz,
            output_dtype,
            order,
            mode,
            cval,
            prefilter,
        )

    semaphore = asyncio.Semaphore(max_concurrent)

    # Order F ensures individual volumes are contiguous in memory
    # Also matches NIfTI, making final save more efficient
    out_array = np.zeros(coordinates.shape[1:] + data.shape[-1:], dtype=output_dtype, order='F')

    tasks = [
        asyncio.create_task(
            worker(
                partial(
                    resample_vol,
                    data=volume,
                    coordinates=coordinates,
                    pe_info=pe_info[volid],
                    jacobian=jacobian,
                    hmc_xfm=hmc_xfms[volid] if hmc_xfms else None,
                    fmap_hz=fmap_hz,
                    output=out_array[..., volid],
                    order=order,
                    mode=mode,
                    cval=cval,
                    prefilter=prefilter,
                ),
                semaphore,
            )
        )
        for volid, volume in enumerate(np.rollaxis(data, -1, 0))
    ]

    await asyncio.gather(*tasks)

    return out_array


def resample_series(
    data: np.ndarray,
    coordinates: np.ndarray,
    pe_info: list[tuple[int, float]],
    jacobian: bool,
    hmc_xfms: list[np.ndarray] | None,
    fmap_hz: np.ndarray,
    output_dtype: np.dtype | None = None,
    order: int = 3,
    mode: str = 'constant',
    cval: float = 0.0,
    prefilter: bool = True,
    nthreads: int = 1,
) -> np.ndarray:
    """Resample a 4D time series at specified coordinates

    This function implements simultaneous head-motion correction and
    susceptibility-distortion correction. It accepts coordinates in
    the source voxel space. It is the responsibility of the caller to
    transform coordinates from any other target space.

    Parameters
    ----------
    data
        The data array to resample
    coordinates
        The first-approximation voxel coordinates to sample from ``data``.
        The first dimension should have length 3.
        The further dimensions determine the shape of the target array.
    pe_info
        A list of readout vectors in the form of (axis, signed-readout-time)
        ``(1, -0.04)`` becomes ``[0, -0.04, 0]``, which indicates that a
        +1 Hz deflection in the field shifts 0.04 voxels toward the start
        of the data array in the second dimension.
    hmc_xfm
        A sequence of affine transformations accounting for head motion from
        the individual volume into the BOLD reference space.
        These affines must be in VOX2VOX form.
    fmap_hz
        The fieldmap, sampled to the target space, in Hz
    output_dtype
        The dtype of the output array.
    order
        Order of interpolation (default: 3 = cubic)
    mode
        How ``data`` is extended beyond its boundaries. See
        :func:`scipy.ndimage.map_coordinates` for more details.
    cval
        Value to fill past edges of ``data`` if ``mode`` is ``'constant'``.
    prefilter
        Determines if ``data`` is pre-filtered before interpolation.
    nthreads
        Number of threads to use for parallel resampling

    Returns
    -------
    resampled_array
        The resampled array, with shape ``coordinates.shape[1:] + (N,)``,
        where N is the number of volumes in ``data``.
    """
    return asyncio.run(
        resample_series_async(
            data=data,
            coordinates=coordinates,
            pe_info=pe_info,
            jacobian=jacobian,
            hmc_xfms=hmc_xfms,
            fmap_hz=fmap_hz,
            output_dtype=output_dtype,
            order=order,
            mode=mode,
            cval=cval,
            prefilter=prefilter,
            max_concurrent=nthreads,
        )
    )


def resample_image(
    source: nb.Nifti1Image,
    target: nb.Nifti1Image,
    transforms: nt.TransformChain,
    fieldmap: nb.Nifti1Image | None,
    pe_info: list[tuple[int, float]] | None,
    jacobian: bool = True,
    nthreads: int = 1,
    output_dtype: np.dtype | str | None = 'f4',
    order: int = 3,
    mode: str = 'constant',
    cval: float = 0.0,
    prefilter: bool = True,
) -> nb.Nifti1Image:
    """Resample a 3- or 4D image into a target space, applying head-motion
    and susceptibility-distortion correction simultaneously.

    Parameters
    ----------
    source
        The 3D bold image or 4D bold series to resample.
    target
        An image sampled in the target space.
    transforms
        A nitransforms TransformChain that maps images from the individual
        BOLD volume space into the target space.
    fieldmap
        The fieldmap, in Hz, sampled in the target space
    pe_info
        A list of readout vectors in the form of (axis, signed-readout-time)
        ``(1, -0.04)`` becomes ``[0, -0.04, 0]``, which indicates that a
        +1 Hz deflection in the field shifts 0.04 voxels toward the start
        of the data array in the second dimension.
    nthreads
        Number of threads to use for parallel resampling
    output_dtype
        The dtype of the output array.
    order
        Order of interpolation (default: 3 = cubic)
    mode
        How ``data`` is extended beyond its boundaries. See
        :func:`scipy.ndimage.map_coordinates` for more details.
    cval
        Value to fill past edges of ``data`` if ``mode`` is ``'constant'``.
    prefilter
        Determines if ``data`` is pre-filtered before interpolation.

    Returns
    -------
    resampled_bold
        The BOLD series resampled into the target space
    """
    if not isinstance(transforms, nt.TransformChain):
        transforms = nt.TransformChain([transforms])
    if isinstance(transforms[-1], nt.linear.LinearTransformsMapping):
        transform_list, hmc = transforms[:-1], transforms[-1]
    else:
        if any(isinstance(xfm, nt.linear.LinearTransformsMapping) for xfm in transforms):
            classes = [xfm.__class__.__name__ for xfm in transforms]
            raise ValueError(f'HMC transforms must come last. Found sequence: {classes}')
        transform_list: list = transforms.transforms
        hmc = []

    # Retrieve the RAS coordinates of the target space
    coordinates = nt.base.SpatialReference.factory(target).ndcoords.astype('f4').T

    # We will operate in voxel space, so get the source affine
    vox2ras = source.affine
    ras2vox = np.linalg.inv(vox2ras)
    # Transform RAS2RAS head motion transforms to VOX2VOX
    hmc_xfms = [ras2vox @ xfm.matrix @ vox2ras for xfm in hmc]

    # After removing the head-motion transforms, add a mapping from boldref
    # world space to voxels. This new transform maps from world coordinates
    # in the target space to voxel coordinates in the source space.
    ref2vox = nt.TransformChain(transform_list + [nt.Affine(ras2vox)])
    mapped_coordinates = ref2vox.map(coordinates)

    # Some identities to reduce special casing downstream
    if fieldmap is None:
        fieldmap = nb.Nifti1Image(np.zeros(target.shape[:3], dtype='f4'), target.affine)
    if pe_info is None:
        pe_info = [[0, 0] for _ in range(source.shape[-1])]

    resampled_data = resample_series(
        data=source.get_fdata(dtype='f4'),
        coordinates=mapped_coordinates.T.reshape((3, *target.shape[:3])),
        pe_info=pe_info,
        jacobian=jacobian,
        hmc_xfms=hmc_xfms,
        fmap_hz=fieldmap.get_fdata(dtype='f4'),
        output_dtype=output_dtype,
        nthreads=nthreads,
        order=order,
        mode=mode,
        cval=cval,
        prefilter=prefilter,
    )
    resampled_img = nb.Nifti1Image(resampled_data, target.affine, target.header)
    resampled_img.set_data_dtype('f4')
    # Preserve zooms of additional dimensions
    resampled_img.header.set_zooms(target.header.get_zooms()[:3] + source.header.get_zooms()[3:])

    return resampled_img


def aligned(aff1: np.ndarray, aff2: np.ndarray) -> bool:
    """Determine if two affines have aligned grids"""
    return np.allclose(
        np.linalg.norm(np.cross(aff1[:-1, :-1].T, aff2[:-1, :-1].T), axis=1),
        0,
        atol=1e-3,
    )


def as_affine(xfm: nt.base.TransformBase) -> nt.Affine | None:
    # Identity transform
    if type(xfm) is nt.base.TransformBase:
        return nt.Affine()

    if isinstance(xfm, nt.Affine):
        return xfm

    if isinstance(xfm, nt.TransformChain) and all(isinstance(x, nt.Affine) for x in xfm):
        return xfm.asaffine()

    return None


def reconstruct_fieldmap(
    coefficients: list[nb.Nifti1Image],
    fmap_reference: nb.Nifti1Image,
    target: nb.Nifti1Image,
    transforms: nt.TransformChain,
) -> nb.Nifti1Image:
    """Resample a fieldmap from B-Spline coefficients into a target space

    If the coefficients and target are aligned, the field is reconstructed
    directly in the target space.
    If not, then the field is reconstructed to the ``fmap_reference``
    resolution, and then resampled according to transforms.

    The former method only applies if the transform chain can be
    collapsed to a single affine transform.

    Parameters
    ----------
    coefficients
        list of B-spline coefficient files. The affine matrices are used
        to reconstruct the knot locations.
    fmap_reference
        The intermediate reference to reconstruct the fieldmap in, if
        it cannot be reconstructed directly in the target space.
    target
        The target space to to resample the fieldmap into.
    transforms
        A nitransforms TransformChain that maps images from the fieldmap
        space into the target space.

    Returns
    -------
    fieldmap
        The fieldmap encoded in ``coefficients``, resampled in the same
        space as ``target``
    """

    direct = False
    affine_xfm = as_affine(transforms)
    if affine_xfm is not None:
        # Transforms maps RAS coordinates in the target to RAS coordinates in
        # the fieldmap space. Composed with target.affine, we have a target voxel
        # to fieldmap RAS affine. Hence, this is projected into fieldmap space.
        projected_affine = affine_xfm.matrix @ target.affine
        # If the coordinates have the same rotation from voxels, we can construct
        # bspline weights efficiently.
        direct = aligned(projected_affine, coefficients[-1].affine)

    if direct:
        reference, _ = ensure_positive_cosines(
            target.__class__(target.dataobj, projected_affine, target.header),
        )
    else:
        if not aligned(fmap_reference.affine, coefficients[-1].affine):
            raise ValueError('Reference passed is not aligned with spline grids')
        reference, _ = ensure_positive_cosines(fmap_reference)

    # Generate tensor-product B-Spline weights
    colmat = sparse_hstack(
        [grid_bspline_weights(reference, level) for level in coefficients]
    ).tocsr()
    coefficients = np.hstack(
        [level.get_fdata(dtype='float32').reshape(-1) for level in coefficients]
    )

    # Reconstruct the fieldmap (in Hz) from coefficients
    fmap_img = nb.Nifti1Image(
        np.reshape(colmat @ coefficients, reference.shape[:3]),
        reference.affine,
    )

    if not direct:
        fmap_img = nt.apply(transforms, fmap_img, reference=target)

    fmap_img.header.set_intent('estimate', name='fieldmap Hz')
    fmap_img.header.set_data_dtype('float32')
    fmap_img.header['cal_max'] = max((abs(fmap_img.dataobj.min()), fmap_img.dataobj.max()))
    fmap_img.header['cal_min'] = -fmap_img.header['cal_max']

    return fmap_img



from pathlib import Path

import nibabel as nb
import nibabel.gifti as ngi
import numpy as np
from nipype.interfaces.base import (
    BaseInterfaceInputSpec,
    File,
    SimpleInterface,
    TraitedSpec,
    traits,
)

from .. import __version__


class _MaskGiftiInputSpec(BaseInterfaceInputSpec):
    in_file = File(exists=True, mandatory=True, desc='Input GIFTI (n-darrays)')
    mask_file = File(exists=True, mandatory=True, desc='Input mask (single binary darray)')
    threshold = traits.Float(
        desc='If mask is probabilistic, inclusion limit',
    )
    metadata = traits.Dict(
        desc='Metadata to insert into GIFTI',
    )


class _MaskGiftiOutputSpec(TraitedSpec):
    out_file = File(desc='Masked file')


class MaskGifti(SimpleInterface):
    """Mask file across GIFTI darrays"""

    input_spec = _MaskGiftiInputSpec
    output_spec = _MaskGiftiOutputSpec

    def _run_interface(self, runtime):
        self._results['out_file'] = _mask_gifti(
            self.inputs.in_file,
            self.inputs.mask_file,
            threshold=self.inputs.threshold or None,
            metadata=self.inputs.metadata,
            newpath=runtime.cwd,
        )
        return runtime


def _mask_gifti(in_file, mask_file, *, threshold=None, metadata=None, newpath=None):
    """
    Mask and create a GIFTI image.
    """
    metadata = metadata or {}

    img = nb.load(in_file)
    mask = nb.load(mask_file).agg_data()

    indices = np.nonzero(mask)[0]
    if threshold is not None:
        indices = np.where(mask > threshold)[0]

    data = img.agg_data()
    if isinstance(data, tuple):
        try:
            data = np.vstack(data)
        except Exception as err:  # noqa: BLE001
            raise NotImplementedError(f'Tricky GIFTI: {in_file} not supported.') from err
    else:
        data = data.T
    masked = data[:, indices]

    # rather than creating new GiftiDataArrays, just modify the data directly
    # and preserve the existing attributes
    for i, darr in enumerate(img.darrays):
        darr.data = masked[i]
        darr.dims = list(masked[i].shape)

    # Finalize by adding additional metadata to file
    metad = {
        **{'CreatedBy': f'MaskGifti (NiBabies-{__version__})'},
        **metadata,
    }
    if int(nb.__version__[0]) >= 4:  # API will change in 4.0.0
        existing_meta = img.meta or {}
        img.meta = ngi.GiftiMetaData({**metad, **existing_meta})
    else:
        meta = img.meta.data or []
        for k, v in metad.items():
            meta.append(ngi.GiftiNVPairs(k, v))
        img.meta.data = meta

    if newpath is None:
        newpath = Path()
    out_file = str((Path(newpath) / f'masked_{Path(in_file).name}').absolute())
    nb.save(img, out_file)
    return out_file



from pathlib import Path

from nireports.assembler.report import Report

from nibabies.data import load as load_data


def run_reports(
    out_dir,
    subject,
    run_uuid,
    session=None,
    out_filename=None,
    reportlets_dir=None,
    packagename=None,
):
    """
    Run the reports.
    """
    return Report(
        out_dir,
        run_uuid,
        subject=subject,
        session=session,
        bootstrap_file=load_data.readable('reports-spec.yml'),
        reportlets_dir=reportlets_dir,
    ).generate_report()


def generate_reports(
    sub_ses_list,
    output_dir,
    run_uuid,
    work_dir=None,
    packagename=None,
):
    """Execute run_reports on a list of subjects."""
    reportlets_dir = None
    if work_dir is not None:
        reportlets_dir = Path(work_dir) / 'reportlets'

    report_errors = []
    for subject, session in sub_ses_list:
        report_errors.append(
            run_reports(
                output_dir,
                subject,
                run_uuid,
                session=session,
                packagename=packagename,
                reportlets_dir=reportlets_dir,
            )
        )

    errno = sum(report_errors)
    if errno:
        import logging

        logger = logging.getLogger('cli')
        error_list = ', '.join(
            f'{subid} ({err})'
            for subid, err in zip(sub_ses_list, report_errors, strict=False)
            if err
        )
        logger.error(
            'Preprocessing did not finish successfully. Errors occurred while processing '
            'data from participants: %s. Check the HTML reports for details.',
            error_list,
        )
    return errno



import shutil
from pathlib import Path

from nipype.interfaces.base import (
    CommandLine,
    CommandLineInputSpec,
    Directory,
    File,
    TraitedSpec,
    traits,
)


class MCRIBReconAllInputSpec(CommandLineInputSpec):
    # Input structure massaging
    outdir = Directory(
        exists=True,
        hash_files=False,
        desc='Path to save output, or path of existing MCRIBS output',
    )
    subjects_dir = Directory(
        exists=True,
        hash_files=False,
        desc='Path to FreeSurfer subjects directory',
    )
    subject_id = traits.Str(
        required=True,
        argstr='%s',
        position=-1,
        desc='Subject ID',
    )
    t1w_file = File(
        exists=True,
        copyfile=True,
        desc='T1w to be used for deformable (must be in register with T2w image)',
    )
    t2w_file = File(
        exists=True,
        copyfile=True,
        desc='T2w (Isotropic + N4 corrected)',
    )
    segmentation_file = File(
        exists=True,
        desc='Segmentation file (skips tissue segmentation)',
    )
    mask_file = File(exists=True, desc='T2w mask')

    # MCRIBS options
    conform = traits.Bool(
        False,
        usedefault=True,
        argstr='--conform',
        desc='Reorients to radiological, axial slice orientation. Resamples to isotropic voxels',
    )
    tissueseg = traits.Bool(
        False,
        usedefault=True,
        argstr='--tissueseg',
        desc='Perform tissue type segmentation',
    )
    surfrecon = traits.Bool(
        False,
        usedefault=True,
        argstr='--surfrecon',
        desc='Reconstruct surfaces',
    )
    surfrecon_method = traits.Enum(
        'Deformable',
        argstr='--surfreconmethod %s',
        requires=['surfrecon'],
        desc='Surface reconstruction method',
    )
    join_thresh = traits.Float(
        argstr='--deformablejointhresh %f',
        requires=['surfrecon'],
        desc='Join threshold parameter for Deformable',
    )
    fast_collision = traits.Bool(
        argstr='--deformablefastcollision',
        requires=['surfrecon'],
        desc='Use Deformable fast collision test',
    )
    autorecon_after_surf = traits.Bool(
        False,
        usedefault=True,
        argstr='--autoreconaftersurf',
        desc='Do all steps after surface reconstruction',
    )
    segstats = traits.Bool(
        argstr='--segstats',
        desc='Compute statistics on segmented volumes',
    )
    nthreads = traits.Int(
        argstr='-nthreads %d',
        desc='Number of threads for multithreading applications',
    )


class MCRIBReconAllOutputSpec(TraitedSpec):
    mcribs_dir = Directory(desc='MCRIBS output directory')
    subjects_dir = Directory(desc='FreeSurfer output directory')


class MCRIBReconAll(CommandLine):
    _cmd = 'MCRIBReconAll'
    input_spec = MCRIBReconAllInputSpec
    output_spec = MCRIBReconAllOutputSpec
    _no_run = False

    _expected_files = {
        'surfrecon': {
            'meshes': (
                'pial-lh-reordered.vtp',
                'pial-rh-reordered.vtp',
                'white-rh.vtp',
                'white-lh.vtp',
            )
        },
        'autorecon': {
            'mri': ('T2.mgz', 'aseg.presurf.mgz', 'ribbon.mgz', 'brain.mgz'),
            'label': ('lh.cortex.label', 'rh.cortex.label'),
            'stats': ('aseg.stats', 'brainvol.stats', 'lh.aparc.stats', 'rh.curv.stats'),
            'surf': (
                'lh.pial', 'rh.pial',
                'lh.white', 'rh.white',
                'lh.curv', 'rh.curv',
                'lh.thickness', 'rh.thickness'),
        }
    }  # fmt:skip

    @property
    def cmdline(self):
        cmd = super().cmdline
        if self.inputs.surfrecon and not self.inputs.t2w_file:
            raise AttributeError('Missing `t2w_file` input.')

        # If an output directory is provided, check if we can skip the run
        if self.inputs.outdir:
            sid = self.inputs.subject_id
            # Check MIRTK surface recon deformable
            if self.inputs.surfrecon:
                surfrecon_dir = Path(self.inputs.outdir) / sid / 'SurfReconDeformable' / sid
                if self._verify_outputs('surfrecon', surfrecon_dir):
                    self._no_run = True
            # Check FS directory population
            if self.inputs.autorecon_after_surf:
                fs_dir = Path(self.inputs.outdir) / sid / 'freesurfer' / sid
                if self._verify_outputs('autorecon', fs_dir):
                    self._no_run = True
                else:
                    self._no_run = False

            if self._no_run:
                return 'echo MCRIBReconAll: nothing to do'
        return cmd

    def _setup_directory_structure(self, mcribs_dir: Path) -> None:
        """
        Create the required structure for skipping steps.

        The directory tree
        ------------------

        <subject_id>/
        ├── RawT2
        │   └── <subject_id>.nii.gz
        ├── SurfReconDeformable
        │   └── <subject_id>
        │       └── temp
        │           └── t2w-image.nii.gz
        ├── TissueSeg
        │   ├── <subject_id>_all_labels.nii.gz
        │   └── <subject_id>_all_labels_manedit.nii.gz
        └── TissueSegDrawEM
            └── <subject_id>
                └── N4
                    └── <subject_id>.nii.gz
        """
        sid = self.inputs.subject_id
        mkdir_kw = {'parents': True, 'exist_ok': True}
        root = mcribs_dir / sid
        root.mkdir(**mkdir_kw)

        # T2w operations
        if self.inputs.t2w_file:
            t2w = root / 'RawT2' / f'{sid}.nii.gz'
            t2w.parent.mkdir(**mkdir_kw)
            if not t2w.exists():
                shutil.copy(self.inputs.t2w_file, str(t2w))

            if not self.inputs.conform:
                t2wiso = root / 'RawT2RadiologicalIsotropic' / f'{sid}.nii.gz'
                t2wiso.parent.mkdir(**mkdir_kw)
                if not t2wiso.exists():
                    t2wiso.symlink_to(f'../RawT2/{sid}.nii.gz')

                n4 = root / 'TissueSegDrawEM' / sid / 'N4' / f'{sid}.nii.gz'
                n4.parent.mkdir(**mkdir_kw)
                if not n4.exists():
                    n4.symlink_to(f'../../../RawT2/{sid}.nii.gz')

        # Segmentation
        if self.inputs.segmentation_file:
            # TissueSeg directive disabled
            tisseg = root / 'TissueSeg' / f'{sid}_all_labels.nii.gz'
            tisseg.parent.mkdir(**mkdir_kw)
            if not tisseg.exists():
                shutil.copy(self.inputs.segmentation_file, str(tisseg))
            manedit = tisseg.parent / f'{sid}_all_labels_manedit.nii.gz'
            if not manedit.exists():
                manedit.symlink_to(tisseg.name)

            if self.inputs.surfrecon:
                t2wseg = root / 'TissueSeg' / f'{sid}_t2w_restore.nii.gz'
                if not t2wseg.exists():
                    t2wseg.symlink_to(f'../RawT2/{sid}.nii.gz')

                surfrec = root / 'SurfReconDeformable' / sid / 'temp' / 't2w-image.nii.gz'
                surfrec.parent.mkdir(**mkdir_kw)
                if not surfrec.exists():
                    surfrec.symlink_to(f'../../../RawT2/{sid}.nii.gz')

                if self.inputs.mask_file:
                    surfrec_mask = surfrec.parent / 'brain-mask.nii.gz'
                    if not surfrec_mask.exists():
                        shutil.copy(self.inputs.mask_file, str(surfrec_mask))

        if self.inputs.surfrecon:
            # Create FreeSurfer layout to safeguard against cd-ing into missing directories
            for d in ('surf', 'mri', 'label', 'scripts', 'stats'):
                (root / 'freesurfer' / sid / d).mkdir(**mkdir_kw)

        # TODO?: T1w -> <subject_id>/RawT1RadiologicalIsotropic/<subjectid>.nii.gz
        return

    def _run_interface(self, runtime):
        # if users wish to preserve their runs
        mcribs_dir = self.inputs.outdir or Path(runtime.cwd) / 'mcribs'
        self._mcribs_dir = Path(mcribs_dir)
        if self.inputs.surfrecon:
            self._setup_directory_structure(self._mcribs_dir)
        # overwrite CWD to be in MCRIB subject's directory
        runtime.cwd = str(self._mcribs_dir / self.inputs.subject_id)
        return super()._run_interface(runtime)

    def _list_outputs(self):
        outputs = self._outputs().get()
        sid = self.inputs.subject_id
        if self.inputs.surfrecon:
            # verify surface reconstruction was successful
            surfrecon_dir = self._mcribs_dir / sid / 'SurfReconDeformable' / sid
            self._verify_outputs('surfrecon', surfrecon_dir, error=True)

        mcribs_fs = self._mcribs_dir / sid / 'freesurfer' / sid
        if self.inputs.autorecon_after_surf:
            self._verify_outputs('autorecon', mcribs_fs, error=True)

        outputs['mcribs_dir'] = str(self._mcribs_dir)
        if self.inputs.autorecon_after_surf and self.inputs.subjects_dir:
            dst = Path(self.inputs.subjects_dir) / self.inputs.subject_id
            if not dst.exists():
                shutil.copytree(mcribs_fs, dst)
                # Create a file to denote this SUBJECTS_DIR was derived from MCRIBS
                logfile = self._mcribs_dir / sid / 'logs' / f'{sid}.log'
                shutil.copyfile(logfile, (dst / 'scripts' / 'mcribs.log'))
            # Copy registration sphere to better match recon-all output
            for hemi in 'lr':
                orig = dst / 'surf' / f'{hemi}h.sphere.reg2'
                symbolic = Path(str(orig)[:-1])
                if orig.exists() and not symbolic.exists():
                    shutil.copyfile(orig, symbolic)
            outputs['subjects_dir'] = self.inputs.subjects_dir

        return outputs

    def _verify_outputs(self, step: str, root: Path, error: bool = False) -> bool:
        """
        Method to check to ensure the expected files are present successful.

        MCRIBReconAll does not return a failing exit code if a step failed, which leads
        this interface to be marked as completed without error in such cases.
        """

        expected = self._expected_files[step]
        for d, files in expected.items():
            for fl in files:
                if not (root / d / fl).exists():
                    if error:
                        raise FileNotFoundError(f'{step.capitalize()} missing: {fl}')
                    return False
        return True



"""Script for testing the subcortical MNI alignment"""
from pathlib import Path


def init_workflow(bold_file, bold_roi, bold_atlas_roi, vol_sigma):
    from nibabies.workflows.bold.alignment import init_subcortical_mni_alignment_wf

    wf = init_subcortical_mni_alignment_wf(vol_sigma=vol_sigma)
    wf.inputs.inputnode.bold_file = bold_file
    wf.inputs.inputnode.bold_roi = bold_roi
    wf.inputs.inputnode.atlas_roi = bold_atlas_roi

    wf.base_dir = Path('workdir').absolute()
    return wf


if __name__ == "__main__":
    from argparse import ArgumentParser, RawTextHelpFormatter

    parser = ArgumentParser(
        description="DCAN subcortical MNI alignment",
        formatter_class=RawTextHelpFormatter,
    )
    parser.add_argument(
        "bold_file",
        type=Path,
        help="the input BOLD file",
    )
    parser.add_argument(
        "bold_roi",
        type=Path,
        help="segmentations in BOLD space",
    )
    parser.add_argument(
        "bold_atlas_roi",
        type=Path,
        help="segmentations in ROI space, unrefined",
    )
    parser.add_argument(
        "--vol-sigma",
        type=float,
        default=0.8,
        help="The sigma for the gaussian volume smoothing kernel, in mm",
    )
    parser.add_argument(
        "--nipype-plugin",
        default="MultiProc",
        help="Nipype plugin to run workflow with",
    )
    opts = parser.parse_args()
    wf = init_workflow(
        opts.bold_file.absolute(),
        opts.bold_roi.absolute(),
        opts.bold_atlas_roi.absolute(),
        vol_sigma=opts.vol_sigma,
    )

    wf.config['execution']['crashfile_format'] = 'txt'
    wf.run(plugin=opts.nipype_plugin)



# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
#
# Copyright 2023 The NiPreps Developers <nipreps@gmail.com>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# We support and encourage derived works from this project, please read
# about our expectations at
#
#     https://www.nipreps.org/community/licensing/
#
"""Check the configuration module and file."""

import os
from unittest.mock import patch

import pytest
from niworkflows.utils.spaces import format_reference
from toml import loads

from nibabies import config
from nibabies.data import load as load_data


def _reset_config():
    """
    Forcibly reload the configuration module to restore defaults.
    .. caution::
      `importlib.reload` creates new sets of objects, but will not remove
      previous references to those objects."""
    import importlib

    importlib.reload(config)


def test_reset_config():
    execution = config.execution
    execution.bids_dir = 'TESTING'
    assert config.execution.bids_dir == 'TESTING'
    _reset_config()
    assert config.execution.bids_dir is None
    # Even though the config module was reset,
    # previous references to config classes
    # have not been touched.
    assert execution.bids_dir == 'TESTING'


def test_config_spaces():
    """Check that all necessary spaces are recorded in the config."""
    settings = loads(load_data.readable('tests/config.toml').read_text())
    for sectionname, configs in settings.items():
        if sectionname != 'environment':
            section = getattr(config, sectionname)
            section.load(configs, init=False)
    config.nipype.init()
    config.loggers.init()
    age = 8
    spaces = _load_spaces(age)
    assert 'MNI152NLin6Asym:res-2' not in [str(s) for s in spaces.get_standard(full_spec=True)]

    assert 'MNI152NLin6Asym_res-2' not in [
        format_reference((s.fullname, s.spec))
        for s in spaces.references
        if s.standard and s.dim == 3
    ]
    # Only enabled if CIFTI is requested
    assert 'MNI152NLin6Asym:res-2' not in [str(s) for s in spaces.get_standard(full_spec=True)]
    assert 'MNI152NLin6Asym_res-2' not in [
        format_reference((s.fullname, s.spec))
        for s in spaces.references
        if s.standard and s.dim == 3
    ]

    config.execution.output_spaces = None

    with pytest.raises(RuntimeError):
        spaces = _load_spaces(None)

    config.execution.output_spaces = None
    config.workflow.cifti_output = '91k'
    spaces = _load_spaces(1)

    assert [str(s) for s in spaces.get_standard(full_spec=True)] == [
        'MNIInfant:cohort-1:res-native',  # Default output space
        'MNI152NLin6Asym:res-2',
        'MNIInfant:cohort-1:res-2',  # CIFTI: MNIInfant (2x2x2) -> MNI152NLin6Asym (2x2x2)
    ]

    assert [
        format_reference((s.fullname, s.spec))
        for s in spaces.references
        if s.standard and s.dim == 3
    ] == ['MNIInfant_cohort-1_res-native', 'MNI152NLin6Asym_res-2', 'MNIInfant_cohort-1_res-2']
    _reset_config()

    config.execution.output_spaces = None
    config.workflow.cifti_output = '170k'
    spaces = _load_spaces(1)

    assert [str(s) for s in spaces.get_standard(full_spec=True)] == [
        'MNIInfant:cohort-1:res-native',  # Default output space
        'MNI152NLin6Asym:res-1',
        'MNIInfant:cohort-1:res-1',
    ]

    assert [
        format_reference((s.fullname, s.spec))
        for s in spaces.references
        if s.standard and s.dim == 3
    ] == ['MNIInfant_cohort-1_res-native', 'MNI152NLin6Asym_res-1', 'MNIInfant_cohort-1_res-1']
    _reset_config()


@pytest.mark.parametrize(
    ('master_seed', 'ants_seed', 'numpy_seed'), [(1, 17612, 8272), (100, 19094, 60232)]
)
def test_prng_seed(master_seed, ants_seed, numpy_seed):
    """Ensure seeds are properly tracked"""
    seeds = config.seeds
    with patch.dict(os.environ, {}):
        seeds.load({'_random_seed': master_seed}, init=True)
        assert seeds.master == master_seed
        assert seeds.ants == ants_seed
        assert seeds.numpy == numpy_seed
        assert os.getenv('ANTS_RANDOM_SEED') == str(ants_seed)

    _reset_config()
    for seed in ('_random_seed', 'master', 'ants', 'numpy'):
        assert getattr(config.seeds, seed) is None


def _load_spaces(age):
    from nibabies.workflows.base import init_execution_spaces, init_workflow_spaces

    # Conditional based on workflow necessities
    spaces = init_workflow_spaces(init_execution_spaces(), age)
    return spaces
